
import operator
import math

from . import classlibrary as clb


clb.ClassLibrary.late_imports(__name__, ('sc3.base.builtins', 'bi'))


class AbstractObject():
    def _compose_unop(self, selector):
        raise NotImplementedError(
            f'{type(self).__name__} does not implement _compose_unop')

    def _compose_binop(self, selector, other):
        raise NotImplementedError(
            f'{type(self).__name__} does not implement _compose_binop')

    def _rcompose_binop(self, selector, other):
        raise NotImplementedError(
            f'{type(self).__name__} does not implement _rcompose_binop')

    def _compose_narop(self, selector, *args):
        raise NotImplementedError(
            f'{type(self).__name__} does not implement _compose_narop')


    # NOTE: If needed by subclasses, operations below, as a whole or in sets,
    # can be implemented as mixins. There are inconsistencies regarding
    # supported operations in specific UGen objects.


    ### Unary operators ###

    def __neg__(self):
        return self._compose_unop(operator.neg)  # -

    def __pos__(self):
        return self._compose_unop(operator.pos)  # + (not in _specialindex)

    def __abs__(self):
        return self._compose_unop(operator.abs)  # abs()

    def __invert__(self):
        return self._compose_unop(operator.invert)  # ~ (bitwise inverse)


    # # Python's numeric type conversion
    #
    # def __complex__(self):
    #     raise NotImplementedError()
    #
    # def __int__(self):
    #     raise NotImplementedError()
    #
    # def __float__(self):  # used by math.floor
    #     raise NotImplementedError()


    # # Python's builtin round and math trunc/floor/ceil.
    #
    # def __round__(self, ndigits=0):
    #     return self._compose_binop(round, ndigits)
    #
    # def __trunc__(self):
    #     return self._compose_unop(math.trunc)  # binary in sclang, possible si problem
    #
    # def __ceil__(self):
    #     return self._compose_unop(math.ceil)
    #
    # def __floor__(self):
    #     return self._compose_unop(math.floor)


    def reciprocal(self):
        return self._compose_unop(bi.reciprocal)

    def ceil(self):
        return self._compose_unop(bi.ceil)

    def floor(self):
        return self._compose_unop(bi.floor)

    def frac(self):
        return self._compose_unop(bi.frac)

    def sign(self):
        return self._compose_unop(bi.sign)

    def log(self):  #, base=math.e):  # In SI is unary, can't change the base.
        return self._compose_unop(bi.log)

    def log2(self):
        return self._compose_unop(bi.log2)

    def log10(self):
        return self._compose_unop(bi.log10)

    # def log1p(self):
    #     return self._compose_unop(bi.log1p)

    def exp(self):
        return self._compose_unop(bi.exp)

    def sin(self):
        return self._compose_unop(bi.sin)

    def cos(self):
        return self._compose_unop(bi.cos)

    def tan(self):
        return self._compose_unop(bi.tan)

    def asin(self):
        return self._compose_unop(bi.asin)

    def acos(self):
        return self._compose_unop(bi.acos)

    def atan(self):
        return self._compose_unop(bi.atan)

    def sinh(self):
        return self._compose_unop(bi.sinh)

    def cosh(self):
        return self._compose_unop(bi.cosh)

    def tanh(self):
        return self._compose_unop(bi.tanh)

    # def asinh(x):
    #     return self._compose_unop(bi.asinh)

    # def acosh(x):
    #     return self._compose_unop(bi.acosh)

    # def atanh(x):
    #     return self._compose_unop(bi.atanh)

    def midicps(self):
        return self._compose_unop(bi.midicps)

    def cpsmidi(self):
        return self._compose_unop(bi.cpsmidi)

    def midiratio(self):
        return self._compose_unop(bi.midiratio)

    def ratiomidi(self):
        return self._compose_unop(bi.ratiomidi)

    def octcps(self):
        return self._compose_unop(bi.octcps)

    def cpsoct(self):
        return self._compose_unop(bi.cpsoct)

    def ampdb(self):
        return self._compose_unop(bi.ampdb)

    def dbamp(self):
        return self._compose_unop(bi.dbamp)

    def squared(self):
        return self._compose_unop(bi.squared)

    def cubed(self):
        return self._compose_unop(bi.cubed)

    def sqrt(self):
        return self._compose_unop(bi.sqrt)

    def rand(self):
        return self._compose_unop(bi.rand)

    def rand2(self):
        return self._compose_unop(bi.rand2)

    def linrand(self):
        return self._compose_unop(bi.linrand)

    def bilinrand(self):
        return self._compose_unop(bi.bilinrand)

    def sum3rand(self):
        return self._compose_unop(bi.sum3rand)

    def coin(self):
        return self._compose_unop(bi.coin)

    def distort(self):
        return self._compose_unop(bi.distort)

    def softclip(self):
        return self._compose_unop(bi.softclip)

    # even
    # odd

    def rectwindow(self):
        return self._compose_unop(bi.rectwindow)

    def hanwindow(self):
        return self._compose_unop(bi.hanwindow)

    def welwindow(self):
        return self._compose_unop(bi.welwindow)

    def triwindow(self):
        return self._compose_unop(bi.triwindow)

    def scurve(self):
        return self._compose_unop(bi.scurve)

    def ramp(self):
        return self._compose_unop(bi.ramp)

    def degrad(self):
        return self._compose_unop(bi.degrad)

    def raddeg(self):
        return self._compose_unop(bi.raddeg)

    # isPositive
    # isNegative
    # isStrictlyPositive

    # complex and cartesian
    # rho
    # theta
    # rotate
    # dist


    ### Binary operators ###

    def __add__(self, other): # +
        return self._compose_binop(operator.add, other)

    def __radd__(self, other):
        return self._rcompose_binop(operator.add, other)

    def __sub__(self, other): # -
        return self._compose_binop(operator.sub, other)

    def __rsub__(self, other):
        return self._rcompose_binop(operator.sub, other)

    def __mul__(self, other): # *
        return self._compose_binop(operator.mul, other)

    def __rmul__(self, other):
        return self._rcompose_binop(operator.mul, other)

    # def __matmul__(self, other): # @
    #     return self._compose_binop(operator.matmul, other)

    # def __rmatmul__(self, other):
    #     return self._rcompose_binop(operator.matmul, other)

    def __truediv__(self, other): # /
        return self._compose_binop(operator.truediv, other)

    def __rtruediv__(self, other):
        return self._rcompose_binop(operator.truediv, other)

    def __floordiv__(self, other): # //
        return self._compose_binop(operator.floordiv, other)

    def __rfloordiv__(self, other):
        return self._rcompose_binop(operator.floordiv, other)

    def __mod__(self, other): # %
        return self._compose_binop(bi.mod, other)

    def __rmod__(self, other):
        return self._rcompose_binop(bi.mod, other)

    # def __divmod__(self, other): # divmod(), método integrado
    #     return self._compose_binop('divmod', other)

    # def __rdivmod__(self, other):
    #     return self._rcompose_binop('divmod', other)

    def __pow__(self, other): # pow(), **, object.__pow__(self, other[, modulo])
        return self._compose_binop(operator.pow, other)

    def __rpow__(self, other):
        return self._rcompose_binop(operator.pow, other)

    def __lshift__(self, other): # <<
        return self._compose_binop(operator.lshift, other)

    def __rlshift__(self, other):
        return self._rcompose_binop(operator.lshift, other)

    def __rshift__(self, other): # >>
        return self._compose_binop(operator.rshift, other)

    def __rrshift__(self, other):
        return self._rcompose_binop(operator.rshift, other)

    def __and__(self, other): # &
        return self._compose_binop(operator.and_, other)

    def __rand__(self, other):
        return self._rcompose_binop(operator.and_, other)

    def __xor__(self, other): # ^
        return self._compose_binop(operator.xor, other)

    def __rxor__(self, other):
        return self._rcompose_binop(operator.xor, other)

    def __or__(self, other): # |
        return self._compose_binop(operator.or_, other)

    def __ror__(self, other):
        return self._rcompose_binop(operator.or_, other)

    # bitHammingDistance  # _HammingDistance -> ./lang/LangPrimSource/PyrBitPrim.cpp

    def __lt__(self, other): # <
        return self._compose_binop(operator.lt, other)

    def __le__(self, other): # <=
        return self._compose_binop(operator.le, other)

    # def __eq__(self, other):  # == (used in ugen dispatch, performBinaryOpOnUGen -> Object.performBinaryOpOnSomething)
    #     return self._compose_binop(operator.eq, other)

    # def __ne__(self, other):  # != # (used in ugen dispatch, performBinaryOpOnUGen -> Object.performBinaryOpOnSomething)
    #     return self._compose_binop(operator.ne, other)

    def __gt__(self, other): # >
        return self._compose_binop(operator.gt, other)

    def __ge__(self, other): # >=
        return self._compose_binop(operator.ge, other)


    def min(self, other):
        return self._compose_binop(bi.min, other)

    def max(self, other=0):
        return self._compose_binop(bi.max, other)

    def lcm(self, other):
        return self._compose_binop(bi.lcm, other)

    def gcd(self, other):
        return self._compose_binop(bi.gcd, other)

    def round(self, other=1):
        return self._compose_binop(bi.round, other)

    def roundup(self, other=1):
        return self._compose_binop(bi.roundup, other)

    def trunc(self, other=1):
        return self._compose_binop(bi.trunc, other)

    def atan2(self, other):
        return self._compose_binop(bi.atan2, other)

    def hypot(self, other):
        return self._compose_binop(bi.hypot, other)

    def hypotx(self, other):
        return self._compose_binop(bi.hypotx, other)

    # def leftshift(self, other):
    #     return self._compose_binop(bi.leftshift, other)
    #
    # def rightshift(self, other):
    #     return self._compose_binop(bi.rightshift, other)
    #
    # def urightshift(self, other):
    #     return self._compose_binop(bi.urightshift, other)

    def ring1(self, other):
        return self._compose_binop(bi.ring1, other)

    def ring2(self, other):
        return self._compose_binop(bi.ring2, other)

    def ring3(self, other):
        return self._compose_binop(bi.ring3, other)

    def ring4(self, other):
        return self._compose_binop(bi.ring4, other)

    def difsqr(self, other):
        return self._compose_binop(bi.difsqr, other)

    def sumsqr(self, other):
        return self._compose_binop(bi.sumsqr, other)

    def sqrsum(self, other):
        return self._compose_binop(bi.sqrsum, other)

    def sqrdif(self, other):
        return self._compose_binop(bi.sqrdif, other)

    def absdif(self, other):
        return self._compose_binop(bi.absdif, other)

    def thresh(self, other):
        return self._compose_binop(bi.thresh, other)

    def amclip(self, other):
        return self._compose_binop(bi.amclip, other)

    def scaleneg(self, other):
        return self._compose_binop(bi.scaleneg, other)

    def clip2(self, other):
        return self._compose_binop(bi.clip2, other)

    def fold2(self, other):
        return self._compose_binop(bi.fold2, other)

    def wrap2(self, other):
        return self._compose_binop(bi.wrap2, other)

    def excess(self, other):
        return self._compose_binop(bi.excess, other)

    def first_arg(self, other):
        # firstArg -> _FirstArg -> SetRaw(a, slotRawObject(a));
        # // this is useful when two ugens need to be called, but only
        # // one of their outputs is needed. (Operators help file).
        return self._compose_binop(bi.first_arg, other)

    def rrand(self, other):
        return self._compose_binop(bi.rrand, other)

    def exprand(self, other):
        return self._compose_binop(bi.exprand, other)


    ### Nary operators ###

    def clip(self, lo, hi):
        return self._compose_narop(bi.clip, lo, hi)

    def wrap(self, lo, hi):
        return self._compose_narop(bi.wrap, lo, hi)

    def fold(self, lo, hi):
        return self._compose_narop(bi.fold, lo, hi)

    def blend(self, other, frac=0.5):
        return self._compose_narop(bi.blend, frac)

    def linlin(self, inmin, inmax, outmin, outmax, clip='minmax'):
        return self._compose_narop(
            bi.linlin, inmin, inmax, outmin, outmax, clip)

    def linexp(self, inmin, inmax, outmin, outmax, clip='minmax'):
        return self._compose_narop(
            bi.linexp, inmin, inmax, outmin, outmax, clip)

    def explin(self, inmin, inmax, outmin, outmax, clip='minmax'):
        return self._compose_narop(
            bi.explin, inmin, inmax, outmin, outmax, clip)

    def expexp(self, inmin, inmax, outmin, outmax, clip='minmax'):
        return self._compose_narop(
            bi.expexp, inmin, inmax, outmin, outmax, clip)

    def lincurve(self, inmin, inmax, outmin, outmax, curve=-4, clip='minmax'):
        return self._compose_narop(
            bi.lincurve, inmin, inmax, outmin, outmax, curve, clip)

    def curvelin(self, inmin, inmax, outmin, outmax, curve=-4, clip='minmax'):
        return self._compose_narop(
            bi.curvelin, inmin, inmax, outmin, outmax, curve, clip)

    def bilin(self, incenter, inmin, inmax, outcenter, outmin, outmax,
              clip='minmax'):
        return self._compose_narop(
            bi.bilin, incenter, inmin, inmax, outcenter, outmin, outmax, clip)

    def biexp(self, incenter, inmin, inmax, outcenter, outmin, outmax,
              clip='minmax'):
        return self._compose_narop(
            bi.biexp, incenter, inmin, inmax, outcenter, outmin, outmax, clip)

    def moddif(self, other, mod=1.0):
        return self._compose_narop(bi.moddif, other, mod)

    # degreeToKey




__version__ = '1.0.0a4'
__all__ = ['base', 'seq', 'synth']


### Library configuration ###

LIB_MODE = 'rt'
LIB_PORT = 57120
LIB_PORT_RANGE = 10
LIB_SETUP_FILE = None


### Configure logger ###

# https://docs.python.org/3/howto/logging-cookbook.html#dealing-with-handlers-that-block

def _init_logger(verbosity, bloking=False):
    import queue
    import logging
    import logging.handlers
    import atexit

    formatter = logging.Formatter('%(levelname)s:%(name)s:%(message)s')

    if bloking:
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        root_logger = logging.getLogger()
        root_logger.addHandler(handler)
    else:
        q = queue.Queue(-1)
        queue_handler = logging.handlers.QueueHandler(q)
        listener_handler = logging.StreamHandler()
        listener_handler.setFormatter(formatter)
        listener = logging.handlers.QueueListener(q, listener_handler)
        root_logger = logging.getLogger()
        root_logger.addHandler(queue_handler)

        listener.start()
        atexit.register(listener.stop)

    root_logger.setLevel(verbosity)

### Init library ###

_libsc3_initialized = False

def init(mode='rt', verbosity='INFO'):
    global _libsc3_initialized

    if _libsc3_initialized:
        return

    mode = mode.lower()
    _init_logger(verbosity, True if mode == 'nrt' else False)

    import sc3.base.main

    if mode == 'rt':
        sc3.base.main.main = sc3.base.main.RtMain
    elif mode == 'nrt':
        sc3.base.main.main = sc3.base.main.NrtMain
    else:
        raise ValueError(f"invalid mode '{mode}'")

    sc3.base.main.main._init()
    _libsc3_initialized = True




import logging

from ...base import utils as utl
from .. import ugen as ugn
from .. import _graphparam as gpp
from . import line as lne


_logger = logging.getLogger(__name__)


class Mix(ugn.PseudoUGen):
    @classmethod
    def new(cls, lst):
        lst = utl.as_list(lst)
        reduced_lst = utl.clump(lst, 4)
        mixed_lst = []
        for item in reduced_lst:
            length = len(item)
            if length == 4:
                mixed_lst.append(ugn.Sum4.new(*item))
            elif length == 3:
                mixed_lst.append(ugn.Sum3.new(*item))
            else:
                mixed_lst.append(utl.list_sum(item))
        if len(mixed_lst) < 3:
            return utl.list_sum(mixed_lst)
        elif len(mixed_lst) == 3:
            return ugn.Sum3.new(*mixed_lst)
        else:
            return cls.new(mixed_lst)

    @classmethod
    def ar(cls, lst):
        result = cls.new(lst)
        rate = result._as_ugen_rate()
        if rate == 'audio':
            return result
        elif rate == 'control':
            return lne.K2A.ar(result)
        elif rate == 'scalar':
            return lne.DC.ar(result)
        else:
            raise ValueError(f'unsupported rate {rate} for Mix.ar')

    @classmethod
    def kr(cls, lst):
        # // 'rate' on an array returns the fastest rate
        # // ('audio' takes precedence over 'control' over 'scalar')
        if gpp.ugen_param(lst)._as_ugen_rate() == 'audio':
            _logger.warning(
                'audio rate input(s) to Mix.kr will '
                'result in signal degradation')
            for i, item in enumerate(lst[:]):
                rate = gpp.ugen_param(item)._as_ugen_rate()
                if rate == 'audio':
                    print(f'{type(item).__name__} {rate}')
                    item._dump_args()
                    lst[i] = lne.A2K.kr(item)
        result = cls.new(lst)
        rate = result._as_ugen_rate()
        if rate == 'control':
            return result
        elif rate == 'scalar':
            return lne.DC.kr(result)
        else:
            raise ValueError(f'unsupported rate {rate} for Mix.kr')


# This is a later added Pseudo UGen used only by JITLib.
# class NumChannels(ugn.PseudoUGen):
#     @classmethod
#     def ar(cls, input, num_channels=2, mixdown=True):
#         input = utl.as_list(input)
#         if len(input) > 1:
#             input = utl.clump(input, bi.ceil(len(input) / num_channels))  # *** BUG: in sclang, will work only if len(input) is >= num_channels AND roundup is multiple of num_channels.
#             result = []
#             for channel in input:
#                 if len(channel) == 1:
#                     result.append(channel[0])
#                 elif mixdown:
#                     result.append(Mix.new(channel))
#                 else:
#                     result.append(channel[0])  # drop the rest
#             if len(result) == 1:
#                 return result[0]
#             else:
#                 return ugn.ChannelList(result)
#         elif num_channels == 1:
#             return input[0]
#         else:
#             return ugn.ChannelList(input * num_channels)  # list dup



"""Line.sc"""

from ...base import builtins as bi
from .. import ugen as ugn
from .. import _graphparam as gpp


class Line(ugn.UGen):
    @classmethod
    def ar(cls, start=0.0, end=1.0, dur=1.0, done_action=0):
        return cls._multi_new('audio', start, end, dur, done_action)

    @classmethod
    def kr(cls, start=0.0, end=1.0, dur=1.0, done_action=0):
        return cls._multi_new('control', start, end, dur, done_action)


class XLine(ugn.UGen):
    @classmethod
    def ar(cls, start=1.0, end=2.0, dur=1.0, done_action=0):
        return cls._multi_new('audio', start, end, dur, done_action)

    @classmethod
    def kr(cls, start=1.0, end=2.0, dur=1.0, done_action=0):
        return cls._multi_new('control', start, end, dur, done_action)


class LinExp(ugn.UGen, ugn.PureUGen):
    @classmethod
    def ar(cls, input=0.0, srclo=0.0, srchi=1.0, dstlo=1.0, dsthi=2.0):
        return cls._multi_new('audio', input, srclo, srchi, dstlo, dsthi)

    @classmethod
    def kr(cls, input=0.0, srclo=0.0, srchi=1.0, dstlo=1.0, dsthi=2.0):
        return cls._multi_new('control', input, srclo, srchi, dstlo, dsthi)

    def _check_inputs(self):
        return self._check_sr_as_first_input()


class LinLin(ugn.PseudoUGen):
    @classmethod
    def ar(cls, input=0.0, srclo=0.0, srchi=1.0, dstlo=1.0, dsthi=2.0):
        scale = (dsthi - dstlo) / (srchi - srclo)
        offset = dstlo - (scale * srclo)
        return ugn.MulAdd.new(input, scale, offset)

    @classmethod
    def kr(cls, input=0.0, srclo=0.0, srchi=1.0, dstlo=1.0, dsthi=2.0):
        scale = (dsthi - dstlo) / (srchi - srclo)
        offset = dstlo - (scale * srclo)
        return input * scale + offset

    @classmethod
    def _method_selector_for_rate(cls, rate):  # FIXME: API, same in SelectX
        if rate == 'audio':
            return 'ar'
        elif rate == 'control':
            return 'kr'
        # return None  # original behaviour
        raise AttributeError(f'{cls.__name__} has no {rate} rate constructor')


class AmpComp(ugn.UGen, ugn.PureUGen):
    @classmethod
    def ar(cls, freq=bi.midicps(60), root=bi.midicps(60), exp=0.3333):
        return cls._multi_new('audio', freq, root, exp)

    @classmethod
    def kr(cls, freq=bi.midicps(60), root=bi.midicps(60), exp=0.3333):
        return cls._multi_new('control', freq, root, exp)

    @classmethod
    def ir(cls, freq=bi.midicps(60), root=bi.midicps(60), exp=0.3333):
        return cls._multi_new('scalar', freq, root, exp)

    def _check_inputs(self):
        if self.rate == 'audio':
            return self._check_sr_as_first_input()
        else:
            return None


class AmpCompA(AmpComp):
    @classmethod
    def ar(cls, freq=1000.0, root=0.0, min_amp=0.32, root_amp=1.0):
        return cls._multi_new('audio', freq, root, min_amp, root_amp)

    @classmethod
    def kr(cls, freq=1000.0, root=0.0, min_amp=0.32, root_amp=1.0):
        return cls._multi_new('control', freq, root, min_amp, root_amp)

    @classmethod
    def ir(cls, freq=1000.0, root=0.0, min_amp=0.32, root_amp=1.0):
        return cls._multi_new('scalar', freq, root, min_amp, root_amp)


class K2A(ugn.UGen, ugn.PureUGen):
    # // Control rate to audio rate converter.
    @classmethod
    def ar(cls, input=0.0):
        return cls._multi_new('audio', input)


class A2K(ugn.UGen, ugn.PureUGen):
    # // Audio rate to control rate converter. only needed in specific cases.
    @classmethod
    def kr(cls, input=0.0):
        return cls._multi_new('control', input)


class T2K(A2K):
    # // Audio rate to control rate trigger converter.
    def _check_inputs(self):
        if gpp.ugen_param(self.inputs[0])._as_ugen_rate() != 'audio':
            return 'First input is not audio rate'
        else:
            return None


class T2A(K2A):
    # // Control rate to audio rate trigger converter.
    @classmethod
    def ar(cls, input=0.0, offset=0):
        return cls._multi_new('audio', input, offset)


class DC(ugn.MultiOutUGen, ugn.PureUGen):
    @classmethod
    def ar(cls, input=0.0):
        return cls._multi_new('audio', input)

    @classmethod
    def kr(cls, input=0.0):
        return cls._multi_new('control', input)

    def _init_ugen(self, *inputs):  # override
        self._inputs = inputs
        return self._init_outputs(len(inputs), self.rate)


class Silent(ugn.PseudoUGen):
    @classmethod
    def ar(cls, num_channels=1):
        sig = DC.ar(0)
        if num_channels == 1:
            return sig
        else:
            return ugn.ChannelList([sig] * num_channels)



"""Global play shortcut mostly for events and lambdas."""

import logging
import inspect

from . import utils as utl
from . import responsedefs as rdf
from ..seq import event as evt
from ..synth import synthdef as sdf
from ..synth import _graphparam as gpp
from ..synth import ugen as ugn
from ..synth import ugens as ugns
from ..synth import envelope as evp
from ..synth import systemdefs as sds
from ..synth import node as nod
from ..synth import buffer as bff


__all__ = ['play']


_logger = logging.getLogger(__name__)


def _play_func(func, target=None, outbus=0, fade=0.01,
               add_action='addToHead', args=()):
    # asDefName.sc
    target = gpp.node_param(target)._as_target()
    server = target.server
    if not server._status_watcher.server_running:
        _logger.warning(f'server {str(server.name)} not running')
        return

    # GraphBuilder.sc
    def wrapper(_iout:'ir'=0):
        # SynthDef.wrap function has to return an UGen-input or
        # None to be chained or not. UGens that don't return a
        # signal should return None. Generated synthdef of this
        # version of func.play() differs a bit from the original.
        result = sdf.SynthDef.wrap(func)
        if result is not None:
            if fade:
                ugns.Control.add_name('gate')
                gate = ugns.Control.kr(1.0)
                result *= ugns.EnvGen.kr(
                    evp.Env.asr(fade, 1, fade, 'lin'), gate, 1, 0, 1, 2)
            result = ugn.SynthObject._replace_zeroes_with_silence(
                utl.as_list(result))
            rate = gpp.ugen_param(result)._as_ugen_rate()
            ugns.Out._multi_new(rate, _iout, *result)

    synthdef = sdf.SynthDef(sds.SystemDefs.generate_tmp_name(), wrapper)
    synth = nod.Synth.basic_new(synthdef.name, server)
    rdf.OscFunc(
        # // Use the /n_end signal to remove the temp synthdef.
        lambda *_: server.send_msg('/d_free', synthdef.name),
        '/n_end', server.addr, arg_template=[synth.node_id]).one_shot()
    args = gpp.node_param(args)._as_control_input()
    synth_msg = synth.new_msg(
        target, ['_iout', outbus, 'out', outbus, *args], add_action)
    synthdef._do_send(server, synth_msg)
    return synth


def _play_buf(buf, loop=False, mul=1, **kwargs):
    if buf._bufnum is None:
        raise bff.BufferAlreadyFreed('_play_buf')

    def buffer_player():
        sig = ugns.PlayBuf.ar(
            buf._channels, buf._bufnum,
            ugns.BufRateScale.kr(buf._bufnum), loop=loop)
        if not loop:
            ugns.FreeSelfWhenDone.kr(sig)
        return sig * mul

    return _play_func(buffer_player, buf._server, **kwargs)


def play(obj=None, *args, **kwargs):
    '''Convenience function to play events and lambdas.

    If called without any argument it will play the default event
    with the default parameters.

    If called only with keyword arguments create an event with
    those keys and plays it.

    If ``obj`` defines the method ``play`` just calls it with
    `*args` and `**kwargs`.

    If ``obj`` is a dict, it creates an event and plays it adding
    kwargs if any. Only keyword arguments are allowed for this case.

    If ``obj`` is a function, it creates a temp SynthDef and makes a
    Synth from it. The function should return the expression to be
    played or have it's own output. This is mostly intended for simple
    lambda functions but can be used as a decorator.

    All versions return the return value of the played object.
    '''

    if not obj and not args and not kwargs:
        # Test tone.
        return evt.event().play()  # TODO: Event is not returning the created server object.
    elif hasattr(obj, 'play'):
        # Is playable.
        return obj.play(*args, **kwargs)
    elif isinstance(obj, dict) and not args:
        # As event.
        return evt.event({**obj, **kwargs}).play()
    elif inspect.isfunction(obj):
        # Decorator or play(lmbd, s, 0, ...).
        return _play_func(obj, *args, **kwargs)
    elif isinstance(obj, bff.Buffer):
        return _play_buf(obj, *args, **kwargs)
    elif obj is None and not args and kwargs:
        # As a keyword only call makes an event.
        return evt.event(kwargs).play()
    else:
        raise ValueError(f'non playable arguments')



"""Thread.sc & Stream.sc part 1"""

from abc import ABC, abstractmethod
import inspect
import enum
import threading
import random
import logging

from . import main as _libsc3
from . import absobject as aob
from . import clock as clk


# NRT uses the name _libsc3 through _init before ClassLibrary initialization.
# OscScore may be created on a different way later. This late import was
# originally defined to reduce cyclic relations and import time when the
# library doesn't need to be completely initialized to write an extension.
# from . import classlibrary as clb
# clb.ClassLibrary.late_imports(__name__, ('sc3.base.main', '_libsc3'))


__all__ = [
    'Routine', 'routine', 'FunctionStream',
    'Condition', 'FlowVar', 'stream', 'embed']


_logger = logging.getLogger(__name__)


### Thread.sc ###


class TimeThread():
    State = enum.Enum('State', [
        'Init', 'Running', 'Suspended', 'Paused', 'Done'])

    def __init__(self, func):
        if not inspect.isfunction(func):
            raise TypeError('TimeThread argument is not a function')

        self.func = func
        self._func_has_inval = (  # Maybe it would be better to require the argument.
            len(inspect.signature(self.func).parameters) > 0)
        self._func_isgenfunc = inspect.isgeneratorfunction(self.func)

        self.parent = None
        self.state = self.State.Init
        self._state_cond = threading.Condition(threading.RLock())
        self._m_seconds = 0.0
        self._clock = None
        self._thread_player = None
        self._rand_seed = None
        self._rgen = _libsc3.main.current_tt._rgen

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self

    # *** TIENE QUE VOLVER A SER PÚBLICO, SE USA MUCHO EN LA DOCUMENTACIÓN,
    # *** RECORDAR POR QUÉ EL ALIAS, CREO PORQUE LO CONVERTÍ EN INTERNO,
    # *** FUE UN ERROR. ANOTAS CON # Updates logical time. LAS LLAMADAS
    # *** A SECONDS (QUE AHORA SON _SECONDS).
    @property
    def seconds(self):
        return self._m_seconds

    @property
    def is_playing(self):
        return (self.state == self.State.Suspended or
                self is _libsc3.main.current_tt)

    @property
    def thread_player(self):
        if self._thread_player is not None:
            return self._thread_player
        else:
            if self.parent is not None\
            and self.parent is not _libsc3.main.main_tt:
                return self.parent.thread_player
            else:
                return self

    @thread_player.setter
    def thread_player(self, player):
        self._thread_player = player

    @property
    def rand_seed(self):
        '''Random seed of the routine.

        By default, Routine's random generators are inherited
        from the parent routine and only change when seeded.
        '''
        return self._rand_seed

    @rand_seed.setter
    def rand_seed(self, x):
        # Routine's rgen are inherited from parent and
        # only changed when seeded. Use sc3.builtins.
        self._rand_seed = x
        self._rgen = random.Random(x)

    @property
    def rand_state(self):
        return self._rgen.getstate()

    @rand_state.setter
    def rand_state(self, data):
        self._rgen.setstate(data)

    # TODO: Maybe for pickling.
    # storeOn { arg stream; stream << "nil"; }
    # archiveAsCompileString { ^true }
    # checkCanArchive { "cannot archive Threads".warn }


class _MainTimeThread(TimeThread):
    def __init__(self):  # override
        self.parent = None
        self.func = None
        self.state = self.State.Init
        self._m_seconds = 0.0
        self._clock = None
        self._thread_player = None

    @property
    def seconds(self):  # override
        # _MainThread set the current physical time when this
        # property is invoked and then spreads to child routines.
        _libsc3.main.update_logical_time()
        return self._m_seconds

    @property
    def is_playing(self):
        return True

    @property
    def _rgen(self):  # override
        return _libsc3.main._m_rgen

    @property
    def rand_seed(self):  # override
        pass

    @rand_seed.setter
    def rand_seed(self, x):  # override
        pass


### Stream.sc part 1 ###


class RoutineException(Exception):
    pass


class StopStream(StopIteration):
    pass


class PausedStream(StopStream):  # Not to be confused with PauseStream which doesn't exists here.
    pass


class YieldAndReset(Exception):
    def __init__(self, yield_value=None):
        super().__init__()  # Doesn't store yield_value in self.args.
        self.yield_value = yield_value


class AlwaysYield(Exception):
    def __init__(self, terminal_value=None):
        super().__init__()  # Doesn't store terminal_value in self.args.
        self.terminal_value = terminal_value


class Stream(aob.AbstractObject, ABC):
    '''
    Streams are iterator-generators compatible objects that implement an
    specific interface to interact with clocks and patterns.
    '''

    ### Iterator protocol ###

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()


    ### Stream protocol ###

    def __stream__(self):
        '''Returns a Stream object.'''
        return self

    def __embed__(self, inval=None):
        '''
        Returns generator-iterator that recursively embeds cointained
        sub-streams.
        '''
        try:
            while True:
                inval = yield self.next(inval)
        except StopStream:
            return inval

    # play

    @abstractmethod
    def next(self, inval=None):
        pass

    def reset(self):
        pass

    # def stream_arg(self): # BUG: Used for Pcollect, Pselect, Preject, method defined by Object and Pattern.
    #     return self

    def all(self, inval=None):
        '''Same as list(stream) but with inval argument.'''
        lst = []
        try:
            while True:
                lst.append(self.next(inval))
        except StopStream:
            pass
        return lst

    # put
    # putN
    # putAll
    # do
    # subSample # Is somewhat specific, evaluates from offset an skipSize amount.
    # generate # Internal method for list comprehension documented in Object.

    # These method could be left for generator list comprehensions.
    # collect
    # reject
    # select

    # dot # // combine item by item with another stream # NOTE: usa FuncStream
    # interlace # // interlace with another stream # NOTE: usa FuncStream
    # ++ (appendStream)
    # append_stream # NOTE: Uses Routine with embedInStream.
    # collate # // ascending order merge of two streams # NOTE: usa interlace
    # <> # Pchain


    ### AbstractObject interface ###

    def _compose_unop(self, selector):
        return UnaryOpStream(selector, self)

    def _compose_binop(self, selector, other):
        return BinaryOpStream(selector, self, stream(other))

    def _rcompose_binop(self, selector, other):
        return BinaryOpStream(selector, stream(other), self)

    def _compose_narop(self, selector, *args):
        args = [stream(x) for x in args]
        return NAryOpStream(selector, self, *args)


    # asEventStreamPlayer
    # trace
    # repeat


### BasicOpStream.sc ###


class UnaryOpStream(Stream):
    def __init__(self, selector, a):
        self.selector = selector
        self.a = a

    def next(self, inval=None):
        a = self.a.next(inval)  # raises StopStream
        return self.selector(a)

    def reset(self):
        self.a.reset()

    # storeOn # TODO


class BinaryOpStream(Stream):
    def __init__(self, selector, a, b):
        self.selector = selector
        self.a = a
        self.b = b

    def next(self, inval=None):
        a = self.a.next(inval)  # raises StopStream
        b = self.b.next(inval)
        return self.selector(a, b)

    def reset(self):
        self.a.reset()
        self.b.reset()

    # storeOn # TODO


# NOTE: See BinaryOpXStream implementation options. Is not possible to
# implemente without a special function operation.


class NAryOpStream(Stream):
    def __init__(self, selector, a, *args):
        self.selector = selector
        self.a = a
        self.args = args  # All args are streams (cast done by _compose_narop).

    def next(self, inval=None):
        a = self.a.next(inval)  # raises StopStream
        args = []
        res = None
        for item in self.args:
            res = item.next(inval)  # raises StopStream
            args.append(res)
        return self.selector(a, *args)

    def reset(self):
        self.a.reset()
        for item in self.args:
            item.reset()

    # storeOn # TODO


class FunctionStream(Stream):  # Was FuncStream
    # Functions could use StopStream as sclang nil but is not nice.
    def __init__(self, next_func, reset_func=None, data=None):
        self.next_func = next_func
        self._next_nargs = len(inspect.signature(self.next_func).parameters)
        self.reset_func = reset_func or (lambda data: None)
        self._reset_nargs = len(inspect.signature(self.reset_func).parameters)
        self.data = data

    def next(self, inval=None):
        # return fn.value(self.next_func, inval, self.data)  # Not cheap.
        # return self.next_func(inval, self.data)  # Mandatory
        if self._next_nargs > 1:
            return self.next_func(inval, self.data)
        elif self._next_nargs > 0:
            return self.next_func(inval)
        else:
            return self.next_func()

    def reset(self):
        # fn.value(self.reset_func, self.data)  # Not cheap.
        # self.reset_func(self.data)  # Mandatory.
        if self._reset_nargs > 0:
            self.reset_func(self.data)
        else:
            self.reset_func()

    # storeArgs


class Routine(TimeThread, Stream):
    class _SENTINEL(): pass

    def __init__(self, func):
        super().__init__(func)
        self._iterator = None
        self._last_value = None
        self._terminal_value = self._SENTINEL

    @classmethod
    def run(cls, func, clock=None, quant=None):
        obj = cls(func)
        obj.play(clock, quant)
        return obj

    def play(self, clock=None, quant=None):
        with self._state_cond:
            if self.state == self.State.Init\
            or self.state == self.state.Paused:
                self.state = self.State.Suspended
                clock = clock or clk.SystemClock
                clock.play(self, quant)
        # Pattern.play return the stream, maybe for API usage constency
        # Stream.play should return self, but I'm not sure.
        # return self

    def next(self, inval=None):
        with self._state_cond:
            if self.state == self.State.Paused:
                raise PausedStream

            # Done & AlwaysYield.
            if self.state == self.State.Done:
                if self._terminal_value is self._SENTINEL:
                    raise StopStream
                else:
                    return self._terminal_value

            self.parent = _libsc3.main.current_tt
            _libsc3.main.current_tt = self
            if self._clock:
                self._m_seconds = self.parent._m_seconds
            else:
                self._m_seconds = self.parent.seconds

            try:
                self.state = self.State.Running
                if self._iterator is None:
                    if self._func_isgenfunc:
                        if self._func_has_inval:
                            self._iterator = self.func(inval)
                        else:
                            self._iterator = self.func()
                        self._last_value = next(self._iterator)
                    else:
                        # Comon functions for routines generate nil streams
                        # in sclang, return value doesn't count. Check over.
                        if self._func_has_inval:
                            # raise AlwaysYield(self.func(inval))
                            self.func(inval)
                        else:
                            # raise AlwaysYield(self.func())
                            self.func()
                        raise AlwaysYield(None)
                else:
                    self._last_value = self._iterator.send(inval)
                self.state = self.State.Suspended
            except StopStream:
                self._iterator = None
                self._last_value = None
                self._clock = None
                self.state = self.State.Done
                raise
            except StopIteration:
                self._iterator = None
                self._last_value = None
                self._clock = None
                self.state = self.State.Done
                raise StopStream from None
            except YieldAndReset as e:
                self._iterator = None
                self.state = self.State.Init
                self._last_value = e.yield_value
            except AlwaysYield as e:
                self._iterator = None
                self._terminal_value = e.terminal_value
                self.state = self.State.Done
                self._last_value = self._terminal_value
            except:
                self.state = self.State.Done  # Failure.
                raise
            finally:
                _libsc3.main.current_tt = self.parent
                self.parent = None

            return self._last_value

    def reset(self):
        with self._state_cond:
            if self.state == self.State.Running:
                raise RoutineException(
                    'cannot be reset within itself except by YieldAndReset')
            else:
                self._iterator = None
                self._clock = None
                self.state = self.State.Init

    def pause(self):
        with self._state_cond:
            if self.state == self.State.Running:
                raise RoutineException('cannot be paused within itself')
            if self.state == self.State.Init\
            or self.state == self.State.Suspended:
                self.state = self.State.Paused

    def resume(self, clock=None, quant=None):
        with self._state_cond:
            if self.state == self.State.Paused:
                self.state = self.State.Suspended
                clock = clock or self._clock or clk.SystemClock
                clock.play(self, quant)

    def stop(self):
        with self._state_cond:
            if self.state == self.State.Running:
                raise RoutineException('cannot be stopped within itself')
            else:
                self._iterator = None
                self._last_value = None
                self._clock = None
                self.state = self.State.Done

    # storeArgs
    # storeOn

    def __awake__(self, beats, seconds, clock):
        return self.next(beats)


# decorator syntax
class routine():
    def __new__(cls, func):
        return Routine(func)

    @staticmethod
    def run(clock=None, quant=None):
        def make_routine(func):
            obj = Routine(func)
            obj.play(clock, quant)
            return obj
        return make_routine


### Condition.sc ###


class Condition():
    def __init__(self, test=False):
        self._test = test
        self._waiting_threads = []

    @property
    def test(self):
        if callable(self._test):
            return self._test()
        else:
            return self._test

    @test.setter
    def test(self, value):
        self._test = value

    def wait(self):
        # This method and hang could be common functions so instead of doing
        # `yielf from condition.wait()` it could be `yield cond.wait()`. That
        # will affect server.sync() for instance. I don't know which is best.
        # However, FlowVar needs to use `yield from` expression.
        if not self.test:
            self._waiting_threads.append(_libsc3.main.current_tt.thread_player)
            yield 'hang'  # Arbitrary non numeric value.
        else:
            yield 0

    def hang(self, value='hang'):
        # // Ignore the test, just wait.
        self._waiting_threads.append(_libsc3.main.current_tt.thread_player)
        yield value

    def signal(self):
        if self.test:
            tmp_wtt = self._waiting_threads
            self._waiting_threads = []
            for tt in tmp_wtt:
                tt._clock.sched(0, tt)

    def unhang(self):
        # // Ignore the test, just resume all waiting threads.
        tmp_wtt = self._waiting_threads
        self._waiting_threads = []
        for tt in tmp_wtt:
            tt._clock.sched(0, tt)


class FlowVar():
    class _UNBOUND(): pass

    def __init__(self):
        self._value = self._UNBOUND
        self.condition = Condition(lambda: self._value is not self._UNBOUND)

    @property
    def value(self):
        yield from self.condition.wait()
        return self._value

    @value.setter
    def value(self, inval):
        if self._value is not self._UNBOUND:
            raise Exception('cannot rebind a FlowVar')
        self._value = inval
        self.condition.signal()


### Type Streams ###


class ValueStream(Stream):
    def __init__(self, value):
        self.value = value


    ### Iterator protocol ###

    def __next__(self):
        return self.value


    ### Stream protocol ###

    def __embed__(self, inval=None):
        # Common objects are infinite streams returning themselves. However,
        # when using embedInStream they become an unique value stream.
        return (yield self.value)

    def next(self, inval=None):
        return self.value


class DictionaryStream(ValueStream):
    ### Iterator protocol ###

    def __next__(self):
        # Object.composeEvent
        return self.value.copy()


    ### Stream protocol ###

    def __embed__(self, indict=None):
        # Dictionary.embedInStream
        # func = self.value.get('embed', None)
        # if func is not None:
        #     yield from func(self.value, indict)
        if indict is None:
            return (yield self.value)
        else:
            indict = indict.copy()
            indict.update(self.value)
            return (yield indict)

    def next(self, indict=None):
        # Event.next
        if indict is None:
            # Object.composeEvent
            return self.value.copy()
        else:
            # Environment.composeEvent
            indict = indict.copy()
            indict.update(self.value)
            return indict


### Module functions ###


def stream(obj):
    '''Converts any object into a Stream.'''

    if hasattr(obj, '__stream__'):
        return obj.__stream__()
    else:
        if isinstance(obj, dict):
            return DictionaryStream(obj)
        else:
            return ValueStream(obj)


def embed(obj, inval=None):
    '''
    Converts any object into a Stream and returns its embeddable form passing
    inval to the `next` calls.
    '''

    if hasattr(obj, '__embed__'):
        return obj.__embed__(inval)
    else:
        if isinstance(obj, dict):
            return DictionaryStream(obj).__embed__(inval)
        else:
            return ValueStream(obj).__embed__(inval)



"""FFT2.sc"""

from .. import ugen as ugn
from . import fftunpacking as ffu


class Convolution(ugn.UGen):
    # // input and kernel are both audio rate changing signals.
    @classmethod
    def ar(cls, input, kernel, frame_size=512):
        return cls._multi_new('audio', input, kernel, frame_size)


class Convolution2(ugn.UGen):
    # // Fixed kernel convolver with fix by nescivi to update
    # // the kernel on receipt of a trigger message.
    @classmethod
    def ar(cls, input, kernel, trigger=0, frame_size=2048):  # NOTE: wrong kernel size crashes the server.
        return cls._multi_new('audio', input, kernel, trigger, frame_size)


class Convolution2L(ugn.UGen):
    # // Fixed kernel convolver with linear crossfade.
    @classmethod
    def ar(cls, input, kernel, trigger=0, frame_size=2048, crossfade=1):
        return cls._multi_new(
            'audio', input, kernel, trigger, frame_size, int(crossfade))


class StereoConvolution2L(ugn.MultiOutUGen):
    # // Fixed kernel stereo convolver with linear crossfade.
    @classmethod
    def ar(cls, input, kernel_L, kernel_R,
           trigger=0, frame_size=2048, crossfade=1):
        return cls._multi_new(
            'audio', input, kernel_L, kernel_R,
            trigger, frame_size, int(crossfade))

    def _init_ugen(self, *inputs):  # override
        self._inputs = inputs
        self._channels = ugn.ChannelList(
            [ugn.OutputProxy.new(self.rate, self, 0),
             ugn.OutputProxy.new(self.rate, self, 1)])
        return self._channels


class Convolution3(ugn.UGen):
    # // Time based convolution by nescivi.
    @classmethod
    def ar(cls, input, kernel, trigger=0, frame_size=2048):
        return cls._multi_new('audio', input, kernel, trigger, frame_size)

    @classmethod
    def kr(cls, input, kernel, trigger=0, frame_size=2048):
        return cls._multi_new('control', input, kernel, trigger, frame_size)


# *** NOTE: CREAR UN ARCHIVO pvugnes.py Y PASAR ESTAS Y LAS DE FFT.sc.
# *** NOTE: TAL VEZ PASAR LAS CONVOLUCIONES A FFT O RENOMBRAR ESTE ARCHIVO
# *** NOTE: A convolution.py, FALTAN PartConv Y OTRAS DE OTROS ARCHIVOS SC.
# *** NOTE: VER QUÉ HACER CON fftunpacking.py QUE DEFINE LA CLAS BASE
# *** NOTE: PV_ChainUGen, FFT ES UN PV_ChainUGen.


class PV_ConformalMap(ffu.PV_ChainUGen):
    # // Sick Lincoln remembers complex analysis courses.
    @classmethod
    def new(cls, buf, areal=0.0, aimag=0.0):
        return cls._multi_new('control', buf, areal, aimag)


class PV_JensenAndersen(ffu.PV_ChainUGen):
    # *** Not sure why this is ar, see _add_to_synth and prefix name.
    # // Jensen andersen inspired FFT feature detector.
    @classmethod
    def ar(cls, buf, propsc=0.25, prophfe=0.25, prophfc=0.25, propsf=0.25,
           threshold=1.0, wait_time=0.04):
        return cls._multi_new('audio', buf, propsc, prophfe, prophfc,
                              propsf, threshold, wait_time)


class PV_HainsworthFoote(ffu.PV_ChainUGen):
    # *** Not sure why this is ar, _add_to_synth and prefix name.
    @classmethod
    def ar(cls, buf, proph=0.0, propf=0.0, threshold=1.0, wait_time=0.04):
        return cls._multi_new('audio', buf, proph, propf, threshold, wait_time)


class RunningSum(ugn.UGen):
    # // Not FFT but useful for time domain onset detection.
    @classmethod
    def ar(cls, input, num_samples=40):
        return cls._multi_new('audio', input, num_samples)

    @classmethod
    def kr(cls, input, num_samples=40):
        return cls._multi_new('control', input, num_samples)

    @classmethod
    def rms(cls, input, num_samples=40):
        sig = RunningSum.ar(input.squared(), num_samples)  # NOTE: si se pasa un valor erróneo a input va a tirar un error distinto al habitual.
        return (sig * (1 / num_samples)).sqrt()



"""Delays.sc"""

from .. import ugen as ugn
from .. import _graphparam as gpp


class Delay1(ugn.UGen, ugn.PureUGen):
    @classmethod
    def ar(cls, input=0.0):
        return cls._multi_new('audio', input)  # NOTE: No gpp.ugen_param(input)._as_audio_rate_input().

    @classmethod
    def kr(cls, input=0.0):
        return cls._multi_new('control', input)


class Delay2(Delay1):
    pass


# // These delays use real time allocated memory.

class DelayN(ugn.UGen, ugn.PureUGen):
    @classmethod
    def ar(cls, input=0.0, max_delay_time=0.2, delay_time=0.2):
        input = gpp.ugen_param(input)._as_audio_rate_input()
        return cls._multi_new('audio', input, max_delay_time, delay_time)

    @classmethod
    def kr(cls, input=0.0, max_delay_time=0.2, delay_time=0.2):
        return cls._multi_new('control', input, max_delay_time, delay_time)


class DelayL(DelayN):
    pass


class DelayC(DelayN):
    pass


class CombN(ugn.UGen, ugn.PureUGen):
    @classmethod
    def ar(cls, input=0.0, max_delay_time=0.2, delay_time=0.2, decay_time=0.2):
        input = gpp.ugen_param(input)._as_audio_rate_input()
        return cls._multi_new(
            'audio', input, max_delay_time, delay_time, decay_time)

    @classmethod
    def kr(cls, input=0.0, max_delay_time=0.2, delay_time=0.2, decay_time=0.2):
        return cls._multi_new(
            'control', input, max_delay_time, delay_time, decay_time)


class CombL(CombN):
    pass


class CombC(CombN):
    pass


class AllpassN(CombN):
    pass


class AllpassL(CombN):
    pass


class AllpassC(CombN):
    pass


# // These delays use shared buffers.

class BufDelayN(ugn.UGen):
    @classmethod
    def ar(cls, buf=0, input=0.0, delay_time=0.2):
        input = gpp.ugen_param(input)._as_audio_rate_input()
        return cls._multi_new('audio', buf, input, delay_time)

    @classmethod
    def kr(cls, buf=0, input=0.0, delay_time=0.2):
        return cls._multi_new('control', buf, input, delay_time)


class BufDelayL(BufDelayN):
    pass


class BufDelayC(BufDelayN):
    pass


class BufCombN(ugn.UGen):
    @classmethod
    def ar(cls, buf=0, input=0.0, delay_time=0.2, decay_time=1.0):
        input = gpp.ugen_param(input)._as_audio_rate_input()
        return cls._multi_new('audio', buf, input, delay_time)


class BufCombL(BufCombN):
    pass


class BufCombC(BufCombN):
    pass


class BufAllpassN(BufCombN):
    pass


class BufAllpassL(BufCombN):
    pass


class BufAllpassC(BufCombN):
    pass


# GrainTap, commented ugen.


class DelTapWr(ugn.UGen):
    @classmethod
    def ar(cls, buf=0, input=0.0):
        input = gpp.ugen_param(input)._as_audio_rate_input()
        return cls._multi_new('audio', buf, input)

    @classmethod
    def kr(cls, buf=0, input=0.0):
        return cls._multi_new('control', buf, input)


class DelTapRd(ugn.UGen):
    @classmethod
    def ar(cls, buf=0, phase=0.0, delay_time=0.2, interp=1):
        return cls._multi_new('audio', buf, phase, delay_time, interp)

    @classmethod
    def kr(cls, buf=0, phase=0.0, delay_time=0.2, interp=1):
        return cls._multi_new('control', buf, phase, delay_time, interp)




import logging
import importlib
import sys


_logger = logging.getLogger(__name__)


class ClassLibrary():
    '''
    This class is a hack to avoid class attribute initialization
    problems caused by very nasty nested cyclic imports. The method
    init() is called at the end of main.
    '''

    _init_list = []
    _imports_list = []
    _initialized = False

    @classmethod
    def add(cls, item, func):
        if cls._initialized:
            func(item)
        else:
            entry = {'item': item, 'func': func}
            cls._init_list.append(entry)

    @classmethod
    def late_imports(cls, module_name, *imports):
        '''
        For imports in cyclic conflict used only at runtime.
        Hack: late imports own imports still can have cyclic conflicts.
        '''
        cls._imports_list.append((module_name, imports))
        if cls._initialized:
            cls._init_imports()

    @classmethod
    def init(cls):
        cls._init_imports()
        while len(cls._init_list) > 0:
            entry = cls._init_list.pop()
            entry['func'](entry['item'])
            _logger.debug('+ init: %s', entry['item'].__name__)
        cls._initialized = True

    @classmethod
    def _init_imports(cls):
        while len(cls._imports_list) > 0:
            name, imports = cls._imports_list.pop()
            for item in imports:
                if isinstance(item, tuple):
                    module = importlib.import_module(item[0])
                    alias = item[1]
                    sys.modules[name].__dict__.update({alias: module})
                else:
                    module = importlib.import_module(item)
                    sys.modules[name].__dict__.update({item: module})




import heapq
import itertools


__all__ = ['TaskQueue']


class TaskQueue():
    """
    This class is an encapsulation of the algorithm found in heapq
    documentation. heapq module in itself use the same principles as
    SuperCollider's clocks implementation. TaskQueue is not thread safe.
    """

    class _REMOVED(): pass

    def __init__(self):
        self._init()

    def _init(self):
        self._queue = []
        self._entry_finder = {}
        self._counter = itertools.count()
        self._removed_counter = 0

    def add(self, prio, task):
        '''Add a new task or update the prio of an existing task.'''
        if task in self._entry_finder:
            self.remove(task)
        count = next(self._counter)
        entry = [prio, count, task]
        self._entry_finder[task] = entry
        heapq.heappush(self._queue, entry)

    def remove(self, task):
        '''Remove an existing task. Does nothing if not found.'''
        try:
            entry = self._entry_finder.pop(task)
            entry[-1] = type(self)._REMOVED
            self._removed_counter += 1
        except KeyError:
            return

    def pop(self):
        '''
        Remove and return the lowest prio entry as a tuple (prio, task).
        Raise KeyError if empty.
        '''
        while self._queue:
            prio, count, task = heapq.heappop(self._queue)
            if task is not type(self)._REMOVED:
                del self._entry_finder[task]
                return (prio, task)
            else:
                self._removed_counter -= 1
        raise KeyError('pop from an empty task queue')

    def peek(self, smallest=True):
        '''
        Return the lowest/highest prio entry as a tuple (prio, task) without
        removing it. Raise KeyError if empty.
        '''
        if self._queue:
            if smallest:
                prio, count, task = heapq.nsmallest(
                    1, self._queue, key=self._small_key)[0]
            else:
                prio, count, task = heapq.nlargest(
                    1, self._queue, key=self._large_key)[0]
            if task is not self._REMOVED:
                return (prio, task)
        raise KeyError('peek from an empty task queue')

    def _small_key(self, item):
        if item[2] is type(self)._REMOVED:
            return [float('inf')] * 2
        else:
            return item[:2]

    def _large_key(self, item):
        if item[2] is type(self)._REMOVED:
            return [float('-inf')] * 2
        else:
            return item[:2]

    def empty(self):
        '''Return True if queue is empty.'''
        return (len(self._queue) - self._removed_counter) == 0

    def clear(self):
        '''Reset the queue to initial state (remove all tasks).'''
        self._init()

    def __iter__(self):
        # FIXME: Returns a generator but creates the whole list first.
        queue = heapq.nsmallest(len(self._queue), self._queue)
        for prio, count, task in queue:
            if task is not type(self)._REMOVED:
                yield (prio, task)

    # def __copy__(self):
    #     ...



"""Trig.sc"""

from .. import ugen as ugn
from .. import _graphparam as gpp
from ...base import utils as utl


class Trig1(ugn.UGen):
    @classmethod
    def ar(cls, input=0.0, dur=0.1):
        return cls._multi_new('audio', input, dur)

    @classmethod
    def kr(cls, input=0.0, dur=0.1):
        return cls._multi_new('control', input, dur)

    @classmethod
    def signal_range(cls):  # override
        return 'unipolar'


class Trig(Trig1):
    pass


class SendTrig(ugn.UGen):
    @classmethod
    def ar(cls, input=0.0, id=0, value=0.0):
        cls._multi_new('audio', input, id, value)
        # return 0.0  # // SendTrig has no output.

    @classmethod
    def kr(cls, input=0.0, id=0, value=0.0):
        cls._multi_new('control', input, id, value)
        # return 0.0  # // SendTrig has no output.

    def _check_inputs(self):  # override
        return self._check_sr_as_first_input()

    def _num_outputs(self):  # override
        return 0

    def _write_output_specs(self, file):  # override
        pass


class SendReply(SendTrig):
    @classmethod
    def ar(cls, trig=0.0, cmd_name='/reply', values=(), reply_id=-1):
        # *** NOTE: values must be a tuple, use a list of tuples for multichannel expansion. Can't be a single value.
        for args in utl.flop([trig, cmd_name, values, reply_id]):
            cls._new1('audio', *args)
        # return 0.0  # // SendReply has no output.

    @classmethod
    def kr(cls, trig=0.0, cmd_name='/reply', values=(), reply_id=-1):
        # *** NOTE: values must be a tuple, use a list of tuples for multichannel expansion. Can't be a single value.
        for args in utl.flop([trig, cmd_name, values, reply_id]):
            cls._new1('control', *args)
        # return 0.0  # // SendReply has no output.

    @classmethod
    def _new1(cls, rate, trig=0.0, cmd_name='/reply', values=None, reply_id=-1):
        cmd_name = [int(x) for x in bytes(cmd_name, 'utf-8')]  # *** TODO: sc ascii method, sclang uses signed vlaues, see Poll.
        return super()._new1(rate, trig, reply_id, len(cmd_name), *cmd_name,
                             *values)


class TDelay(Trig1):
    def _check_inputs(self):  # override
        return self._check_sr_as_first_input()


class Latch(ugn.UGen):
    @classmethod
    def ar(cls, input=0.0, trig=0.0):
        return cls._muli_new('audio', input, trig)

    @classmethod
    def kr(cls, input=0.0, trig=0.0):
        return cls._muli_new('control', input, trig)


class Gate(Latch):
    pass


class PulseCount(ugn.UGen):
    @classmethod
    def ar(cls, trig=0.0, reset=0.0):
        return cls._multi_new('audio', trig, reset)

    @classmethod
    def kr(cls, trig=0.0, reset=0.0):
        return cls._multi_new('control', trig, reset)

    def _check_inputs(self):  # override
        return self._check_sr_as_first_input()


class Peak(ugn.UGen):
    @classmethod
    def ar(cls, input=0.0, trig=0.0):
        return cls._muli_new('audio', input, trig)

    @classmethod
    def kr(cls, input=0.0, trig=0.0):
        return cls._muli_new('control', input, trig)

    def _check_inputs(self):  # override
        if self.rate == 'control'\
        and gpp.ugen_param(self.inputs[0])._as_ugen_rate() == 'audio':
            return self._check_valid_inputs()
        else:
            return self._check_sr_as_first_input()


class RunningMin(Peak):
    pass


class RunningMax(Peak):
    pass


class Stepper(ugn.UGen):
    @classmethod
    def ar(cls, trig=0, reset=0, min=0, max=7, step=1, resetval=None):
        if resetval is None:
            resetval = min
        return cls._multi_new('audio', trig, reset, min, max, step, resetval)

    @classmethod
    def kr(cls, trig=0, reset=0, min=0, max=7, step=1, resetval=None):
        if resetval is None:
            resetval = min
        return cls._multi_new('control', trig, reset, min, max, step, resetval)

    def _check_inputs(self):  # override
        return self._check_sr_as_first_input()


class PulseDivider(ugn.UGen):
    @classmethod
    def ar(cls, trig=0.0, div=2.0, start=0.0):
        return cls._multi_new('audio', trig, div, start)

    @classmethod
    def kr(cls, trig=0.0, div=2.0, start=0.0):
        return cls._multi_new('control', trig, div, start)


class SetResetFF(PulseCount):
    pass


class ToggleFF(ugn.UGen):
    @classmethod
    def ar(cls, trig=0.0):
        return cls._multi_new('audio', trig)

    @classmethod
    def kr(cls, trig=0.0):
        return cls._multi_new('control', trig)


class ZeroCrossing(ugn.UGen):
    @classmethod
    def ar(cls, input=0.0):
        return cls._multi_new('audio', input)

    @classmethod
    def kr(cls, input=0.0):
        return cls._multi_new('control', input)

    def _check_inputs(self):  # override
        return self._check_sr_as_first_input()


class Timer(ugn.UGen):
    # // Output is the time between two triggers.
    @classmethod
    def ar(cls, trig=0.0):
        return cls._multi_new('audio', trig)

    @classmethod
    def kr(cls, trig=0.0):
        return cls._multi_new('control', trig)

    def _check_inputs(self):  # override
        return self._check_sr_as_first_input()


class Sweep(ugn.UGen):
    # // Output sweeps up in value at rate per
    # // second the trigger resets to zero.
    @classmethod
    def ar(cls, trig=0.0, rate=1.0):
        return cls._multi_new('audio', trig, rate)

    @classmethod
    def kr(cls, trig=0.0, rate=1.0):
        return cls._multi_new('control', trig, rate)


class Phasor(ugn.UGen):
    @classmethod
    def ar(cls, trig=0.0, rate=1.0, start=0.0, end=1.0, reset_pos=0.0):
        return cls._multi_new('audio', trig, rate, start, end, reset_pos)

    @classmethod
    def kr(cls, trig=0.0, rate=1.0, start=0.0, end=1.0, reset_pos=0.0):
        return cls._multi_new('control', trig, rate, start, end, reset_pos)


class PeakFollower(ugn.UGen):
    @classmethod
    def ar(cls, input=0.0, decay=0.999):
        return cls._multi_new('audio', input, decay)

    @classmethod
    def kr(cls, input=0.0, rate=0.999):
        return cls._multi_new('control', input, decay)


class Pitch(ugn.MultiOutUGen):
    _default_rate = 'control'

    @classmethod
    def kr(cls, input=0.0, init_freq=440.0, min_freq=60.0, max_freq=4000.0,
           exec_freq=100.0, max_bins_per_octave=16, median=1,
           amp_threshold=0.01, peak_threshold=0.5, down_sample=1, clar=0):
        return cls._multi_new('control', input, init_freq, min_freq, max_freq,
                              exec_freq, max_bins_per_octave, median,
                              amp_threshold, peak_threshold, down_sample, clar)

    def _init_ugen(self, *inputs):  # override
        self._inputs = inputs
        return self._init_outputs(2, self.rate)


class InRange(ugn.UGen):
    @classmethod
    def ar(cls, input=0.0, lo=0.0, hi=1.0):
        return cls._multi_new('audio', input, lo, hi)

    @classmethod
    def kr(cls, input=0.0, lo=0.0, hi=1.0):
        return cls._multi_new('control', input, lo, hi)

    @classmethod
    def ir(cls, input=0.0, lo=0.0, hi=1.0):
        return cls._multi_new('scalar', input, lo, hi)


class InRect(ugn.UGen):
    @classmethod
    def ar(cls, x=0.0, y=0.0, rect=(0.0, 0.0, 0.0, 0.0)):  # left, top, right, bottom (x, y, x, y lines on screen)
        return cls._multi_new('audio', x, y, *rect)

    @classmethod
    def kr(cls, x=0.0, y=0.0, rect=(0.0, 0.0, 0.0, 0.0)):  # left, top, right, bottom (x, y, x, y lines on screen)
        return cls._multi_new('control', x, y, *rect)


# Trapezoid, commented UGen.


class Fold(InRange):
    pass


class Clip(InRange):
    pass


class Wrap(InRange):
    pass


class Schmidt(InRange):
    pass


class ModDif(ugn.UGen):
    @classmethod
    def ar(cls, x=0.0, y=0.0, mod=1.0):
        return cls._multi_new('audio', x, y, mod)

    @classmethod
    def kr(cls, x=0.0, y=0.0, mod=1.0):
        return cls._multi_new('control', x, y, mod)

    @classmethod
    def ir(cls, x=0.0, y=0.0, mod=1.0):
        return cls._multi_new('scalar', x, y, mod)


class MostChange(ugn.UGen):
    @classmethod
    def ar(cls, a=0.0, b=0.0):
        return cls._multi_new('audio', a, b)

    @classmethod
    def kr(cls, a=0.0, b=0.0):
        return cls._multi_new('control', a, b)


class LeastChange(MostChange):
    pass


class LastValue(ugn.UGen):
    @classmethod
    def ar(cls, input=0.0, diff=0.01):
        return cls._multi_new('audio', input, diff)

    @classmethod
    def kr(cls, input=0.0, diff=0.01):
        return cls._multi_new('control', input, diff)


class SendPeakRMS(ugn.UGen):
    @classmethod
    def ar(cls, sig, reply_rate=20.0, peak_lag=3, cmd_name='/reply',
           reply_id=-1):
        return cls._new1('audio', utl.as_list(sig), reply_rate, peak_lag,
                         cmd_name, reply_id)

    @classmethod
    def kr(cls, sig, reply_rate=20.0, peak_lag=3, cmd_name='/reply',
           reply_id=-1):
        return cls._new1('control', utl.as_list(sig), reply_rate, peak_lag,
                         cmd_name, reply_id)

    @classmethod
    def _new1(cls, rate, sig, reply_rate, peak_lag, cmd_name, reply_id):
        cmd_name = [int(x) for x in bytes(cmd_name, 'utf-8')]  # *** TODO: sc ascii method, sclang uses signed vlaues, see Poll.
        return super()._new1(rate, reply_rate, peak_lag, reply_id, len(sig),
                             *utl.flatten(sig), len(cmd_name), *cmd_name)

    def _num_outputs(self):  # override
        return 0

    def _write_output_specs(self, file):  # override
        pass



"""
Import related hooks.
"""

import importlib as _importlib
import pkgutil as _pkgutil
import inspect as _inspect
import sys as _sys


def import_all_module(name, package=None, *, bind):
    '''
    Imports all non internal clases and make the available from the
    module passed to bind. The module referred by bind must be already
    available in sys.modules dict. Return value is a mapping of the form
    {'class_name': class_obj, ...}. Parameters name and package are the
    arguments of importlib.import_module().
    '''
    module = _importlib.import_module(name, package)
    mapping = dict(_inspect.getmembers(module, _inspect.isclass))
    _sys.modules[bind].__dict__.update(mapping)
    return mapping


def import_all_package(path, name, *, bind):
    '''
    Same as import_all_module but imports all non internal classes from all
    non internal modules within a package and makes them available from the
    module passed to bind.
    '''
    ret = dict()
    for module_info in _pkgutil.walk_packages(path, name + '.'):
        if module_info.name.split('.')[-1][0] == '_':
            continue
        module = _importlib.import_module(module_info.name)
        mapping = dict(_inspect.getmembers(module, _inspect.isclass))
        _sys.modules[bind].__dict__.update(mapping)
        ret.update(mapping)
    return ret



"""
The content of this file is highly experimental.

A Patch is a language-side control graph as a possible replacement for
SuperCollider's event streams. It organize patterns and actions scheduling
played by Routine as a synchronous graph that resembles synth graphs.

Objects, operands and operations, within a patch form a graph that is evaluated
cyclically by triggers.

The main difference with event streams is that any object can be triggered
separately and thus have its own time function. Triggers can be combined to
create superimposed functions in time, e.g. each pattern can have its own
timing and target the same output object.

It's inspired by the patterns library of SuperCollider and Max/PD control flow.
However, its implementation differs from them and it creates its own set of
rules and behaviour, it's a different mix.

Sequential programming was preferred to simplify side-effect actions such as
resource instantiation and cleanup.
"""

import logging
import collections
import itertools
import sys
import traceback

from ..base import clock as clk
from ..base import _taskq as tsq
from ..base import stream as stm
from ..base import absobject as aob
from ..synth import node as nod
from ..synth import server as srv


_logger = logging.getLogger(__name__)


class _UniqueList(list):
    def append(self, item):
        if item not in self:
            super().append(item)

    def extend(self, iterable):
        for item in iterable:
            super().append(item)

    def remove(self, item):
        if item in self:
            super().remove(item)


class Patch():
    _Entry = collections.namedtuple(
        '_Entry', ['beat', 'next_beat', 'trig', 'messages', 'roots'])
    current_patch = None

    def __init__(self):
        self._parent = Patch.current_patch
        self._clock = None
        self._outlet = None
        self._roots = _UniqueList()
        self._triggers = _UniqueList()
        self._messages = _UniqueList()
        self._cleaners = _UniqueList()
        self._beat = 0.0
        self._cycle = 0
        self._queue = None
        self._neatq = None
        self._routine = None
        self.__stop = False
        self._tempo_scale = 1.0

    @property
    def outlet(self):
        return self._outlet

    @outlet.setter
    def outlet(self, value):
        if self._outlet:
            raise Exception('Patch can only have one Outlet object.')
        self._outlet = value

    @property
    def roots(self):
        return tuple(self._roots)

    def play(self, clock=None, quant=None):
        if self._routine:
            return

        def patch_routine():
            yield from self._gen_function()

        if self._parent:
            if clock:
                _logger.warning("Sub-patches inherit parent's clock.")
            self._clock = self._parent._clock
        else:
            self._clock = clock or clk.SystemClock
        self._routine = stm.Routine(patch_routine)
        self._routine.play(self._clock, quant)  # SystemClock ignores quant.
        # *** TODO: Routine could notify, or not.

    def stop(self):
        if self._routine:
            # self._routine.stop()
            self.__stop = True  # *** check CmdPeriod interaction.
            self._routine = None  # Only once.

    def _init_queue(self):
        # Evaluación ordenada de los triggers y las roots.
        self._queue = tsq.TaskQueue()
        for root in self._roots:
            for trigger in root._get_triggers():
                self._add_trigger(trigger)
            for message in root._get_messages():
                self._add_message(message)

        self._neatq = tsq.TaskQueue()
        for neatobj in self._cleaners:
            self._neatq.add(neatobj.delay, neatobj)

    def _add_trigger(self, trigger):
        if trigger in self._triggers or not trigger._active:
            return
        self._triggers.append(trigger)
        messages = trigger._get_active_messages()
        roots = trigger._get_active_roots()
        self._queue.add(
            self._beat + next(trigger), (trigger, messages, roots))

    def _remove_trigger(self, trigger):
        if trigger not in self._triggers:
            return
        if not trigger._get_active_messages()\
        and len(trigger._get_active_roots()) < 2:
            trigger._active = False  # Cancel in queue.
            self._triggers.remove(trigger)

    def _add_message(self, message):
        if message._active:
            self._messages.append(message)

    def _remove_message(self, message):
        self._messages.remove(message)

    def _gen_function(self):
        self._init_queue()

        try:
            # Initial RootBox evaluation, self._cycle == 0.
            # Messages are always evaluated before roots within a cycle.
            self._evaluate_cycle(self._messages + self._roots)
        except StopIteration:
            return

        beat = 0
        prev_beat = 0

        while not self._queue.empty():
            # Cycle data.

            evaluables = []
            beat, (trigger, messages, roots) = self._queue.pop()

            # round should sync sub-patches so they are evaluated always first
            # at the same clock's queue time because play schedules them before.
            yield round((beat - prev_beat) * self._tempo_scale, 9)
            if self.__stop:
                break
            self._beat = beat
            self._cycle += 1

            # Triggers are evaluated first each cycle (after yield).
            next_beat = next(trigger)  # Triggers are infinite.
            evaluables.append(self._Entry(
                beat, next_beat, trigger, messages, roots))

            while not self._queue.empty()\
            and round(beat, 9) == round(self._queue.peek()[0], 9):  # Sincroniza pero introduce un error diferente, hay que ver si converge para el delta de cada trigger.
                trigger, messages, roots = self._queue.pop()[1]
                next_beat = next(trigger)
                evaluables.append(self._Entry(
                    beat, next_beat, trigger, messages, roots))

            # Evaluation.

            messages = _UniqueList()
            roots = _UniqueList()
            for entry in evaluables:
                messages.extend(entry.messages)
                roots.extend(entry.roots)

            try:
                self._evaluate_cycle(messages + roots)
            except StopIteration:
                break

            for entry in evaluables:
                if entry.trig._active:
                    messages = entry.trig._get_active_messages()
                    roots = entry.trig._get_active_roots()
                    if not messages and not roots:
                        continue
                    # Time tends to error/overflow by resolution over time.
                    self._queue.add(
                        entry.beat + entry.next_beat,
                        (entry.trig, messages, roots))

            prev_beat = beat

        # Cleanup

        if self._neatq.empty():
            return

        # concatenate with last beat delta.
        prev_delay = beat - prev_beat
        if prev_delay > self._neatq.peek()[0]:
            prev_delay = self._neatq.peek()[0]

        while not self._neatq.empty():
            delay, neatobj = self._neatq.pop()
            yield (delay - prev_delay) * self._tempo_scale
            try:
                prev_patch = Patch.current_patch
                Patch.current_patch = self
                neatobj._evaluate()
            finally:
                Patch.current_patch = prev_patch
            prev_delay = delay

    def _evaluate_cycle(self, evaluables):
        try:
            # Patch puede ser context.
            exception = False
            prev_patch = Patch.current_patch
            Patch.current_patch = self
            for out in evaluables:
                try:
                    if out._active:  # Messages deactivate RootBox in its last iteration that is one more for rootboxes.
                        out._evaluate()  # *** Also has internal catch and raise.
                    else:
                        exception = True
                except StopIteration:
                    exception = True
            if exception:
                if not any(r._active for r in self._messages + self._roots):
                    raise StopIteration
        finally:
            Patch.current_patch = prev_patch


class PatchFunction():
    def __init__(self, func):
        self.func = func

    def __call__(self, *args, play=True, clock=None, quant=None, **kwargs):
        try:
            # Patch puede ser context.
            new_patch = Patch()
            prev_patch = Patch.current_patch
            Patch.current_patch = new_patch
            self.func(*args, **kwargs)
        finally:
            Patch.current_patch = prev_patch
        if play:
            new_patch.play(clock, quant)
        return new_patch


# Decorator syntax.
def patch(func):
    return PatchFunction(func)


'''
from sc3.all import *
from sc3.seq.patch import *

@patch
def test():
    seq1 = Seq([1, 2, 3])
    seq2 = Seq([10, 20, 30, 40], tgg=Trig(1))
    seq3 = Seq([100, 200, 300, 400], tgg=Trig(3))

    res1 = seq1 + seq2
    res2 = (1000 + seq1) + seq2 + seq3

    Trace(res1, 'res1')
    Trace(res2, 'res2')

t = test(play=False)
print([out for out in t.roots])
t.play()
'''


class BoxObject():
    class __NOCACHE(): pass

    def __init__(self, tgg=None, msg=None):
        self._active = True
        self._patch = Patch.current_patch
        self._prev_cycle = -1
        self._cache = self.__NOCACHE
        self._parents = _UniqueList()
        self._children = _UniqueList()
        self._roots = _UniqueList()
        self._messages = _UniqueList()
        self._triggers = _UniqueList()
        if tgg:
            tgg._connect(self)
        if msg:
            msg._connect(self)

    def __iter__(self):
        return self

    def __next__(self):
        raise NotImplementedError(f'{type(self).__name__}.__next__')

    def _evaluate(self):
        # Patch is a generator function that creates an timed generator
        # iterator. BoxObjects are evaluated by cycle. A cycle is started
        # by any Trig contained in the Patch. BoxObject with triggers
        # are evaluated with it's own Trig's timing by cleaning the cache.
        # BoxObject without a Trig is evaluated by the cycle of
        # something's else Trig if is in its op branch.
        # As consecuencie, if a BoxObject without Trig is in the branch
        # of more than other BoxObject with different Trigs it will be
        # consumed by the triggers of every shared expression, a copy whould
        # be needed to avoid this. Is it too much compliated?

        if not self._active:
            raise StopIteration

        try:
            if self._triggers:
                if self._cached:
                    return self._cache
                else:
                    ret = self._cache = next(self)
                    return ret
            else:
                if self._patch._cycle > self._prev_cycle:
                    self._prev_cycle = self._patch._cycle
                    ret = self._cache = next(self)
                    return ret
                else:
                    return self._cache
        except StopIteration:
            self._deactivate()
            raise

    def _deactivate(self):
        # StopIteration deactivates parent in series.
        # Exceptions are managed in __next__.
        if self._active:
            self._active = False

    @property
    def _cached(self):
        return self._cache is not self.__NOCACHE

    def _clear_cache(self):
        self._cache = self.__NOCACHE
        for r in self._parents:
            if r._cached:
                r._clear_cache()

    def _add_parent(self, box, dyn=False):
        self._parents.append(box)
        box._children.append(self)
        if isinstance(box, RootBox) and not box._replaced:
            self._roots.append(box)
        if dyn:
            for trigger in self._get_triggers():
                self._patch._add_trigger(trigger)
            for message in self._get_messages():
                self._patch._add_message(message)

    def _remove_parent(self, box, dyn=False):
        if dyn:
            for trigger in self._get_triggers():
                self._patch._remove_trigger(trigger)
            for message in self._get_messages():
                self._patch._remove_message(message)
        self._parents.remove(box)
        box._children.remove(self)
        if isinstance(box, RootBox):
            self._roots.remove(box)
            if not box._children:
                self._patch._roots.remove(box)

    def _get_roots(self):
        # Outputs are searched towards the roots.
        ret = _UniqueList()
        for p in self._parents:
            for r in p._get_roots():
                ret.append(r)
        for r in self._roots:
            ret.append(r)
        return ret

    def _get_triggers(self):
        # Triggers are searched towards the leaves.
        ret = _UniqueList()
        for child in self._children:
            ret.extend(child._get_triggers())
        for message in self._messages:
            ret.extend(message._triggers)
        if self._triggers:
            ret.extend(self._triggers)
        return ret

    def _get_triggered_objects(self):
        ret = _UniqueList()
        for child in self._children:
            ret.extend(child._get_triggered_objects())
        if self._triggers:
            ret.append(self)
        for message in self._messages:
            if message._trigger:
                ret.append(message)
        return ret

    def _get_msg_recv(self):
        return self

    def _get_messages(self):
        ret = _UniqueList()
        for child in self._children:
            ret.extend(child._get_messages())
        if self._messages:
            ret.extend(self._messages)
        return ret


class TriggerObject():
    '''
    Triggers are iterators that just return floats as deltas.
    They are not part of the graph as nodes, they are transversal.
    '''
    def __init__(self):
        self._iterator = None
        self._objs = []
        self._active = True

    def __iter__(self):
        return self

    def __next__(self):
        for obj in self._objs:
            obj._clear_cache()
        return next(self._iterator)

    def _connect(self, obj):
        if not obj in self._objs:
            self._objs.append(obj)
            obj._triggers.append(self)

    def _disconnect(self, obj):  # no useful?
        if obj in self._objs:
            self._objs.remove(obj)
            obj._triggers.remove(self)

    @property
    def _boxes(self):
        return [o for o in self._objs if isinstance(o, BoxObject)]

    @property
    def _messages(self):
        return [o for o in self._objs if isinstance(o, Message)]

    def _get_active_roots(self):
        roots = set(r for b in self._boxes for r in b._get_roots() if r._active)
        roots |= set(
            r for m in self._messages for o in m._objs\
            for r in o._get_roots() if r._active)
        return tuple(roots)

    def _get_active_messages(self):
        return tuple(m for m in self._messages if m._active)


class Trig(TriggerObject):
    def __init__(self, hz=1):
        super().__init__()
        self._iterator = itertools.repeat(1.0 / hz)


class Every(TriggerObject):
    def __init__(self, time=1):
        super().__init__()
        if isinstance(time, (list, tuple)):
            self._iterator = itertools.cycle(time)
        else:
            self._iterator = itertools.repeat(time)


class Within(TriggerObject):
    def __init__(self, time=1, n=1):
        super().__init__()
        if isinstance(n, (list, tuple)):
            self._iterator = itertools.cycle(
                itertools.chain(*[[time / i] * i for i in n]))
        else:
            self._iterator = itertools.repeat(time / n)


'''
from sc3.all import *
from sc3.seq.patch import *

@patch
def test():
    seq = Seq(range(20), tgg=Within(1, [4, 3, 2, 1]))
    Trace(seq)

p = test()
'''


class RootBox(BoxObject):
    def __init__(self, tgg=None, msg=None):
        super().__init__(tgg, msg)
        self._roots.append(self)  # Needed for triggers.
        self._patch._roots.append(self)
        self._replaced = False

    def _add_parent(self, box, dyn=False):
        # Roots are replaced by other roots.
        if isinstance(box, RootBox):
            self._remove_from_roots()
        super()._add_parent(box, dyn)

    def _remove_from_roots(self):
        self._roots.remove(self)
        for child in self._children:
            child._roots.remove(self)
        self._patch._roots.remove(self)
        self._replaced = True

    # def _deactivate(self):
    #     # Is redundant _evaluate_cycle checks not only active triggers but
    #     # also active roots.
    #     self._active = False
    #     for trigger in self._triggers:
    #         # No other active root for this root's trigger.
    #         if not any(r._active for o in trigger._objs for r in o._get_roots()):
    #             trigger._active = False


class Outlet(RootBox):
    def __init__(self, value, tgg=None):
        super().__init__(tgg)
        self._patch.outlet = self
        if isinstance(value, (list, tuple)):
            self._value = ValueList(value)
        elif not isinstance(value, ValueList):
            self._value = ValueList([value])
        self._value._add_parent(self)

    def __next__(self):
        return self._value._evaluate()

    def __getitem__(self, index):
        return self._value[index]

    def __iter__(self):
        # As iterable behaves different.
        return iter(self._value)

    def __len__(self):
        return len(self._value)


class ValueList(BoxObject):
    def __init__(self, lst):
        super().__init__()
        self._lst = []
        self._len = len(lst)
        for obj in lst:
            obj = Value(obj)
            obj._add_parent(self)
            self._lst.append(obj)

    def __next__(self):
        ret = []
        ended = 0
        for value in self._lst:
            try:
                ret.append(value._evaluate())
            except StopIteration:
                ret.append(None)
                ended += 1
        if ended == self._len:
            raise StopIteration
        return ret

    def __getitem__(self, index):
        return self._lst[index]

    def __iter__(self):
        # As iterable behaves different.
        return iter(self._lst)

    def __len__(self):
        return self._len


'''
# - Tengo que cambiar la implementación, que cada outlet sea independiente y
#   si se usan todos juntos funciona como ahora (corta el primero que termina),
#   pero que esto sea explícito al crear las inlets (ahí se ponene en ValueList,
#   si se obtienen por separado no y cada una termina cuando termina).
from sc3.all import *
from sc3.seq.patch import *

@patch
def outlst():
    a = Seq([1, 2, 3, 4], tgg=Trig(1))
    b = Seq([10, 20, 30, 40], tgg=Trig(2))
    c = Seq([100, 200, 300, 400], tgg=Trig(3))
    o = Outlet([a, b, c])
    # Trace(o[0])
    # a, b, c = o
    # Trace(ValueList([a, b, c]))

@patch
def inlst():
    # a = Inlet(outlst())
    # Trace(a, tgg=Trig(3))

    # a = Inlet(outlst(), 0)
    # Trace(a, tgg=Trig(3))

    # lst = Inlet(outlst(), slice(2))
    # Trace(lst, tgg=Trig(3))

    # a = Inlet(outlst())[0]
    # Trace(a, tgg=Trig(3))

    # a, b, c = Inlet(outlst())
    # Trace(ValueList([a, b, c]), tgg=Trig(3))

    lst = Inlet(outlst())
    Trace(ValueList([*lst]), tgg=Trig(3))

# outlst()
inlst()
'''


class Event(RootBox):

    class _EventDelta(TriggerObject):
        def __init__(self, delta):
            super().__init__()
            self._delta = delta

        def __next__(self):
            for obj in self._objs:
                obj._clear_cache()
            self._active = False
            return self._delta

    def __init__(self, delta, obj):
        super().__init__(tgg=self._EventDelta(delta))
        self._delta = delta
        self._obj = Value(obj)
        # Needs to remove because is going to replace it later.
        if isinstance(self._obj, RootBox):
            self._obj._remove_from_roots()
        for message in self._obj._get_messages():
            message._active = False
        self._wait = True

    def _add_object(self):
        # Messages also need initial evaluation. In this
        # case it will be inside a root evaluation call.
        messages = self._obj._get_messages()
        messages = [m for m in messages if m not in self._patch._messages]
        for m in messages: m._active = True
        self._patch._evaluate_cycle(messages)
        self._obj._add_parent(self, True)

    def __next__(self):
        if self._wait:
            if self._delta > self._patch._beat:
                return
            self._wait = False
            self._add_object()
        return self._obj._evaluate()


'''
from sc3.all import *
from sc3.seq.patch import *

@patch
def p1():
    a = Event(3, Seq([1, 2, 3, 4, 5], tgg=Trig(4)))
    Trace(a, tgg=Trig(1))

p = p1()
'''

'''
from sc3.all import *
from sc3.seq.patch import *

s.boot()

@synthdef
def ping(freq=440, amp=0.05):
    sig = SinOsc(freq) * amp
    env = EnvGen.kr(Env.perc(), done_action=Done.FREE_SELF)
    Out(0, (sig * env).dup())

@patch
def test():
    Event(2, Note(name='ping', freq=Seq([440, 550, 660], repeat=5, tgg=Trig(1.1))))
    Event(4, Note(name='ping', freq=Seq([1220, 1230, 1320], repeat=4, tgg=Trig(1))))
    Event(6, Note(name='ping', freq=Seq([350, 500, 750], repeat=5, tgg=Trig(3))))

# p = test()
'''

'''
from logging import info
from sc3.all import *
from sc3.seq.patch import *

class FakeObject():
    def on(self, note, vel=63):
        info(f'note on! {note}, {vel}')
    def off(self, note, vel=63):
        info(f'note off! {note}, {vel}')

@patch
def test():
    msg = Message(['on 60 16', 'on 72', 'off 60', 'off 72'], tgg=Trig(1))
    box = Box(FakeObject(), msg=msg)
    e = Event(5, box)
    Trace(e)

p = test()
'''


class Trace(RootBox):
    def __init__(self, graph, prefix=None, tgg=None, msg=None):
        super().__init__(tgg, msg)
        self._graph = Value(graph)
        self._prefix = prefix or 'Trace'
        self._graph._add_parent(self)

    def __next__(self):
        value = self._graph._evaluate()
        _logger.info(
            f'{self._prefix}: <{type(self._graph).__name__}, '
            f'{hex(id(self._graph))}>, cycle: {self._patch._cycle}, '
            f'value: {value}')
        return value


'''
from sc3.all import *
from sc3.seq.patch import *

@patch
def test():
    a = Seq([1, 2, 3], tgg=Trig(1))
    b = Seq([10, 20, 30, 40], tgg=Trig(1))
    c = Value(100, Trig(3))
    r = a + b + c
    Trace(r)

test()
'''


class Tempo(RootBox):
    def __init__(self, bpm, tgg=None, msg=None):
        super().__init__(tgg, msg)
        self._bpm = bpm
        self._hz = Value(bpm / 60)
        self._hz._add_parent(self)

    def __next__(self):
        value = self._hz._evaluate()
        self._patch._tempo_scale = 1.0 / value
        return value


'''
from sc3.all import *
from sc3.seq.patch import *

@patch
def test():
    Tempo(Seq([60, 120, 60, 120, 60]), tgg=Every(4))
    # Tempo(Seq([60, 120] * 3), tgg=Every([4, 3]))
    seq1 = Seq(range(20))
    Trace(seq1, tgg=Trig(1))

p = test()
'''


class Note(RootBox):
    # noteEvent of sc Event.

    # *** consider linear and real polyrhythms (melodic and harmonic).
    # *** composed trigs as one iterator (&, ||, ??)?
    # *** Note(dur=Every(0.5))?

    def __init__(self, *args, **kwargs):
        super().__init__()
        self._params = dict()
        for i, v in enumerate(args, 0):
            v = Value(v)
            v._add_parent(self)
            self._params[i] = v
        for k, v in kwargs.items():
            v = Value(v)
            v._add_parent(self)
            self._params[k] = v

    def __next__(self):
        ...  # bundle
        params = {k: v._evaluate() for k, v in self._params.items()}
        def_name = params.pop('name', 'default')
        target = params.pop('target', None)
        add_action = params.pop('add_action', 'addToHead')
        register = params.pop('register', None)
        args = [i for t in params.items() for i in t]
        synth = nod.Synth(def_name, args, target, add_action, register)
        ... # release en base a dur msg.
        ... # bundle
        ... # send
        return synth


'''
from sc3.all import *
from sc3.seq.patch import *

s.boot()

@synthdef
def ping(freq=440, amp=0.1):
    sig = SinOsc(freq) * amp
    env = EnvGen.kr(Env.perc(), done_action=Done.FREE_SELF)
    Out(0, (sig * env).dup())

@patch
def test():
    freq = Seq([440, 480, 540, 580] * 2, tgg=Trig(1))
    amp = Seq([0.01, 0.1] * 4)  #, tgg=Trig(0.4))
    note = Note(name='ping', freq=freq, amp=amp)
    # Trig(3)._connect(note)

# p = test()
'''


class Inlet(BoxObject):
    def __init__(self, patch, index=None):
        super().__init__()
        self._input_patch = patch
        self._input = patch.outlet
        self._index = index if index is not None else slice(len(patch.outlet))

    def __next__(self):
        if self._input and self._input._active:
            return self._input._cache[self._index]
        else:
            raise StopIteration

    def __getitem__(self, index):
        return type(self)(self._input_patch, index)

    def __iter__(self):
        # As iterable behaves different.
        return (type(self)(self._input_patch, i) for i in range(len(self)))

    def __len__(self):
        return len(self._input)


'''
from sc3.all import *
from sc3.seq.patch import *

@patch
def a():
    freq = Seq([1, 2, 3, 4], tgg=Trig(3))
    Trace(freq, 'Seq A')
    Outlet(freq)

@patch
def b():
    pa = a()

    freq = Inlet(pa)
    Trace(freq, 'Inlet', Trig(1))

    freq2 = Seq([10, 20, 30])
    Trace(freq2, 'Seq B', Trig(1))

pb = b()
'''


class Box(BoxObject):  # *** TEST
    def __init__(self, obj, tgg=None, msg=None):
        super().__init__(tgg, msg)
        self._obj = obj

    def __next__(self):
        return self._obj

    def _get_msg_recv(self):
        return self._obj


class Message():
    def __init__(self, lst, tgg, bang=True):
        self._active = True
        self._lst = lst
        self.__iterator = iter(lst)
        self._triggers = _UniqueList()
        tgg._connect(self)
        self._bang = bang
        self._objs = []

    def __iter__(self):
        return self

    def __next__(self):
        next_msg = next(self.__iterator)
        next_msg = self._parse(next_msg)
        for obj in self._objs:
            if self._bang:
                obj._clear_cache()
            recv = obj._get_msg_recv()
            getattr(recv, next_msg[0])(*next_msg[1:])  # *** AttributeError
        return next_msg

    def _parse(self, msg):
        # ['selector 1 "2" 3', 'selector 3 2.1']
        # [('selector', 1, '2', 3), ('selector', 3, 2.1)]
        if isinstance(msg, str):
            msg = msg.split()  # *** BUG: can pass valid characters that form expressions: "60," is tuple, [1,2,3], etc.
            for i, v in enumerate(msg[1:][:], 1):
                msg[i] = eval(v, dict())  # *** NameError a log.
        return msg

    def _connect(self, obj):
        if not obj in self._objs:
            self._objs.append(obj)
            obj._messages.append(self)

    def _disconnect(self, obj):  # no useful?
        if obj in self._objs:
            self._objs.remove(obj)
            obj._messages.remove(self)


    # Needed by _evaluate_cycle.

    def _evaluate(self):
        try:
            return next(self)
        except StopIteration:
            self._deactivate()
            raise

    def _deactivate(self):
        self._active = False
        for trigger in self._triggers:
            # Disable trigger if doesn't have other connection.
            if len(trigger._objs) == 1:
                trigger._active = False
            # Disable rootbox if doesn't have other active triggers.
            for root in trigger._get_active_roots():
                if not any(t._active for t in root._get_triggers()):
                    root._active = False


    # Needed by triggers interface.

    def _clear_cache(self):
        pass


'''
from logging import info  # to avoid a very annoying ipython bug.
from sc3.all import *
from sc3.seq.patch import *

class FakeObject():
    def on(self, note, vel=63):
        info(f'note on! {note}, {vel}')
    def off(self, note, vel=63):
        info(f'note off! {note}, {vel}')

@patch
def test():
    msg = Message(['on 60 16', 'on 72', 'off 60', 'off 72'], tgg=Trig(1))
    box = Box(FakeObject(), msg=msg)  # Put it in a box.
    Trace(box)  # Outlet(box)

p = test()
'''


class Tidyner():
    def __init__(self):
        self._patch = Patch.current_patch
        self._patch._cleaners.append(self)


class Cleanup(Tidyner):
    def __init__(self, lst, method=None, delay=None):
        super().__init__()
        method = method or 'free'
        delay = 1.0 if delay is None else delay
        self.lst = []
        for item in lst:
            if isinstance(item, tuple):
                self.lst.append(item)  # (obj, 'method', arg1, arg2, ...)
            else:
                self.lst.append((item, method))  # (obj, method)
        self.method = method
        self.delay = delay

    def _evaluate(self):
        for obj, method, *args in self.lst:
            try:
                getattr(obj, method)(*args)
            except:
                _logger.error(
                    ''.join(traceback.format_exception(
                        *sys.exc_info(), -1)))


class CleanupFunction(Tidyner):
    def __init__(self, func, args=None, delay=None):
        super().__init__()
        self.func = func
        self.args = () if args is None else args
        self.delay = 1.0 if delay is None else delay

    def _evaluate(self):
        try:
            self.func(*self.args)
        except:
            _logger.error(
                ''.join(traceback.format_exception(
                    *sys.exc_info(), -1)))


# Decorator syntax.
def cleanup(func=None, *, args=(), delay=None):
    if func is None and delay is not None:
        def _(func):
            return CleanupFunction(func, args, delay)
        return _
    else:
        return CleanupFunction(func)


'''
from sc3.all import *
from sc3.seq.patch import *

@patch
def test():
    Trace(Seq([1, 2, 3]), tgg=Trig(1))

    Cleanup([('hola', 'split', 'o')], delay=2)
    Cleanup([('hola', 'split', 'o')], delay=1)

    @cleanup(delay=1.5)
    def tidyner():
        print('neat!')

p = test()
'''

'''
from logging import info
from sc3.all import *
from sc3.seq.patch import *

s.boot()

@synthdef
def ping(freq=440, amp=0.05):
    sig = SinOsc(freq) * amp
    env = EnvGen.kr(Env.perc(0.2), done_action=Done.FREE_SELF)
    Out(0, (sig * env).dup())

@patch
def test():
    group = Group()
    freq = Seq([60, 62, 64] * 20, tgg=Trig(5)).midicps()
    scale = Seq([1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7] * 40, tgg=Trig(7.1))
    Note(name='ping', freq=freq * scale, target=group)

    @cleanup(delay=2)
    def tidyner():
        info('free group')
        group.free()

# p = test()
'''

'''
from sc3.all import *
from sc3.seq.patch import *

s.boot()

@synthdef
def ping(freq=440, amp=0.05):
    sig = SinOsc(freq) * amp
    env = EnvGen.kr(Env.perc(), done_action=Done.FREE_SELF)
    Out(0, (sig * env).dup())

@patch
def test():
    g = Group()
    h = Group(g, add_action='addAfter')
    seq = Seq([0, 2, 4, 5, 7, 9, 11], repeat=100, tgg=Every(1))
    target = If(seq < 5, g, h)
    Note(name='ping', freq=bi.midicps(seq + 60), target=target)
    Cleanup([g, h])

# p = test()
'''


class AbstractBox(BoxObject, aob.AbstractObject):
    def _compose_unop(self, selector):
        return UnopBox(selector, self)

    def _compose_binop(self, selector, other):
        return BinopBox(selector, self, other)

    def _rcompose_binop(self, selector, other):
        return BinopBox(selector, other, self)

    def _compose_narop(self, selector, *args):
        return NaropBox(selector, self, *args)


class UnopBox(AbstractBox):
    def __init__(self, selector, a):
        super().__init__()
        self.selector = selector
        self.a = Value(a)
        a._add_parent(self)

    def __next__(self):
        return self.selector(self.a._evaluate())


class BinopBox(AbstractBox):
    def __init__(self, selector, a, b):
        super().__init__()
        self.selector = selector
        self.a = Value(a)
        self.b = Value(b)
        for obj in self.a, self.b:
            obj._add_parent(self)

    def __next__(self):
        a = self.a._evaluate()
        b = self.b._evaluate()
        return self.selector(a, b)


class NaropBox(AbstractBox):
    def __init__(self, selector, a, *args):
        super().__init__()
        self.selector = selector
        self.a = Value(a)
        a._add_parent(self)
        self.args = []
        for obj in args:
            obj = Value(obj)
            obj._add_parent(self)
            self.args.append(obj)

    def __next__(self):
        args = [obj._evaluate() for obj in self.args]
        return self.selector(self.a._evaluate(), *args)


class If(AbstractBox):
    def __init__(self, cond, true, false):
        super().__init__()
        self.cond = Value(cond)
        true = Value(true)
        false = Value(false)
        self._check_fork(true, false)
        self.fork = (true, false)
        for obj in (self.cond, *self.fork):  # inactive branch keeps running its own triggers.
            obj._add_parent(self)

    def _check_fork(self, *fork):
        for b in fork:
            if (isinstance(b, Outlet) or hasattr(b, '_get_roots'))\
            and b._get_roots():
                raise ValueError("true/false expressions can't contain roots")

    def __next__(self):
        cond = self.cond._evaluate()
        cond = int(not cond)
        return self.fork[cond]._evaluate()


'''
from sc3.all import *
from sc3.seq.patch import *

@patch
def test():
    a = Seq([1, 2, 3, 4], tgg=Trig(1))
    b = 0
    c = a + b
    i = If(c > 2, True, False)
    Trace(i)

g = test(play=False)._gen_function()
[value for value in g]
'''

'''
from sc3.all import *
from sc3.seq.patch import *

@patch
def test():
    seq1 = Seq(range(20), tgg=Trig(4))
    seq2 = Seq(range(0, 100, 10), tgg=Trig(4))
    seq3 = Seq(range(0, 1000, 100), tgg=Trig(4))
    res = If(seq1 % 2, seq3, seq2)
    Trace(res)

p = test()
'''


class MetaValue(type):
    def __call__(cls, value, tgg=None):
        if isinstance(value, (TriggerObject, Message)):  #, RootBox)):
            raise TypeError(f'{type(value).__name__} is not valid input')
        if isinstance(value, BoxObject):
            return value
        obj = cls.__new__(cls, value, tgg)
        obj.__init__(value, tgg)
        return obj


class Value(AbstractBox, metaclass=MetaValue):
    def __init__(self, value, tgg=None):
        super().__init__(tgg)
        self._value = value

    def __next__(self):
        return self._value


'''
from sc3.all import *
from sc3.seq.patch import *

@patch
def test():
    a = Value(1)
    b = Value(2)
    c = a + b
    Trace(c, tgg=Trig(1))

g = test(play=False)._gen_function()
for _ in range(10): next(g)
'''


class Map(BoxObject):
    def __init__(self, dct, tgg=None):
        super().__init__(tgg)
        self._params = dict()
        for key, value in dct.items():
            value = Value(value)
            value._add_parent(self)
            self._params[key] = value

    def __next__(self):
        return {k: v._evaluate() for k, v in self._params.items()}


'''
from sc3.all import *
from sc3.seq.patch import *

@patch
def t():
    m = Map({
        'instr': 'trombón',
        'freq': Seq([100, 200, 300], tgg=Every(2)),
        'amp': Value(0.1, tgg=Every(0.5))
    })
    Trace(m)

p = t()
'''


class Seq(AbstractBox):
    def __init__(self, lst, repeat=1, tgg=None):
        super().__init__(tgg)
        self._lst = lst
        self._len = len(lst)
        self.__iterator = self._seq_iterator()
        self._repeat = repeat

    def _seq_iterator(self):
        for _ in range(self._repeat):
            for obj in self._lst:
                if isinstance(obj, BoxObject):
                    try:
                        obj._add_parent(self, True)
                        while True:
                             yield obj._evaluate()
                    except StopIteration:
                        pass
                    obj._remove_parent(self, True)
                else:
                    yield obj

    def __next__(self):
        return next(self.__iterator)

    def __len__(self):
        return self._len


'''
from sc3.all import *
from sc3.seq.patch import *

@patch
def p1():
    a = Seq([
        Seq([1, 2], tgg=Trig(1)),
        Seq([10, 20], tgg=Trig(2)),
        Seq([1000, 2000], tgg=Trig(1))
    ], tgg=Trig(4))
    Trace(a)  #, tgg=Trig(1))

p = p1()
'''

'''
from sc3.all import *
from sc3.seq.patch import *

@patch
def p1():
    a = Seq([
        Seq([1, 2], tgg=Trig(1)),
        11, 22,  # Special case, no trigger in the chain, stops after 11.
        Seq([1000, 2000], tgg=Trig(1))
    ])
    Trace(a)  #, tgg=Trig(1))

p = p1()
'''


class FunctionBox(AbstractBox):
    def __init__(self, func, *args, tgg=None, **kwargs):
        super().__init__(tgg)
        self.func = func
        self.args = [Value(arg) for arg in args]
        self.kwargs = {key: Value(value) for key, value in kwargs.items()}
        for obj in (*self.args, *self.kwargs.values()):
            obj._add_parent(self)

    def __next__(self):
        args = [x._evaluate() for x in self.args]
        kwargs = {key: value._evaluate() for key, value in self.kwargs.items()}
        return self.func(*args, **kwargs)



"""Model.sc"""

import weakref

from . import functions as fn


__all__ = ['NotificationCenter']


class NotificationCenter():
    _registrations = weakref.WeakKeyDictionary()

    def __new__(cls):
        return cls

    @classmethod
    def notify(cls, obj, msg, *args, **kwargs):
        if obj in cls._registrations and msg in cls._registrations[obj]:
            for listener, action in cls._registrations[obj][msg].copy().items():
                fn.value(action, obj, msg, listener, *args, **kwargs)

    @classmethod
    def register(cls, obj, msg, listener, action):
        if obj not in cls._registrations:
            cls._registrations[obj] = dict()
        if msg not in cls._registrations[obj]:
            cls._registrations[obj][msg] = weakref.WeakKeyDictionary()
        cls._registrations[obj][msg][listener] = action

    @classmethod
    def unregister(cls, obj, msg=None, listener=None):
        err = False
        try:
            if msg is None:
                del cls._registrations[obj]
            elif listener is None:
                del cls._registrations[obj][msg]
            else:
                del cls._registrations[obj][msg][listener]
        except KeyError as e:
            err = True
        if err:
            raise KeyError(
                f'no registration found for ({obj}, {msg}, {listener})')

    @classmethod
    def register_one_shot(cls, obj, msg, listener, action):
        def one_shot_action(*args):
            action(*args)
            cls.unregister(obj, msg, listener)
        cls.register(obj, msg, listener, one_shot_action)

    @classmethod
    def registration_exists(cls, obj, msg, listener):
        if obj in cls._registrations\
        and msg in cls._registrations[obj]\
        and listener in cls._registrations[obj][msg]:
            return True
        else:
            return False

    @classmethod
    def clear(cls):
        cls._registrations = weakref.WeakKeyDictionary()



"""DiskIO.sc"""

from ...base import utils as utl
from .. import ugen as ugn
from .. import _graphparam as gpp


class DiskOut(ugn.UGen):
    @classmethod
    def ar(cls, buf, input):
        return cls._multi_new('audio', buf, *utl.as_list(input))

    def _check_inputs(self):  # override
        if self._rate == 'audio':
            for input in self._inputs[1:]:
                if gpp.ugen_param(input)._as_ugen_rate() != 'audio':
                    return f'input was not audio rate: {input}'
        return self._check_valid_inputs()


class DiskIn(ugn.MultiOutUGen):
    @classmethod
    def ar(cls, num_channels, buf, loop=0):
        return cls._multi_new('audio', num_channels, buf, loop)

    def _init_ugen(self, num_channels, *inputs):  # override
        self._inputs = inputs
        return self._init_outputs(num_channels, self.rate)


class VDiskIn(ugn.MultiOutUGen):
    @classmethod
    def ar(cls, num_channels, buf, rate=1, loop=0, send_id=0):
        return cls._multi_new('audio', num_channels, buf, rate, loop, send_id)

    def _init_ugen(self, num_channels, *inputs):  # override
        self._inputs = inputs
        return self._init_outputs(num_channels, self.rate)
